# -*- coding: utf-8 -*-
import logging

from flask import g, request
from flask_restful import Api
from http import HTTPStatus

from vantage6.common import logger_name
from vantage6.server.permission import (
    PermissionManager,
    Scope as S,
    Operation as P
)
from vantage6.server.resource import (
    with_node,
    ServicesResources
)
from vantage6.server import db
from vantage6.server.resource.common.output_schema import PortSchema
from vantage6.server.resource.common.input_schema import PortInputSchema
from vantage6.server.model import (
    Run,
    AlgorithmPort,
)
from vantage6.server.resource import with_container

module_name = logger_name(__name__)
log = logging.getLogger(module_name)


def setup(api: Api, api_base: str, services: dict) -> None:
    """
    Setup the port resource.

    Parameters
    ----------
    api : Api
        Flask restful api instance
    api_base : str
        Base url of the api
    services : dict
        Dictionary with services required for the resource endpoints
    """
    path = "/".join([api_base, module_name])
    log.info(f'Setting up "{path}" and subdirectories')

    api.add_resource(
        Ports,
        path,
        endpoint='port_without_id',
        methods=('POST', 'DELETE'),
        resource_class_kwargs=services
    )
    api.add_resource(
        VPNAddress,
        api_base + '/vpn/algorithm/addresses',
        endpoint='vpn_address',
        methods=('GET',),
        resource_class_kwargs=services
    )


# Schemas
port_schema = PortSchema()
port_input_schema = PortInputSchema()


# -----------------------------------------------------------------------------
# Permissions
# -----------------------------------------------------------------------------
def permissions(permissions: PermissionManager) -> None:
    """
    Define the permissions for this resource.

    Parameters
    ----------
    permissions : PermissionManager
        Permission manager instance to which permissions are added
    """
    add = permissions.appender(module_name)

    add(scope=S.GLOBAL, operation=P.VIEW, description="view any port")
    add(scope=S.ORGANIZATION, operation=P.VIEW, assign_to_container=True,
        assign_to_node=True, description="view ports of your organizations "
        "collaborations")


# ------------------------------------------------------------------------------
# Resources / API's
# ------------------------------------------------------------------------------
class PortBase(ServicesResources):
    def __init__(self, socketio, mail, api, permissions, config):
        super().__init__(socketio, mail, api, permissions, config)
        self.r = getattr(self.permissions, module_name)


class Ports(PortBase):
    @with_node
    def post(self):
        """Create a list of port description
        ---
        description: >-
          Creates a description of a port that is available for VPN
          communication for a certain algorithm. Only the node on which the
          algorithm is running is allowed to create this.\n

          This endpoint is not accessible for users, but only for
          authenticated nodes.

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  port:
                    type: integer
                    description: Port number that receives container's VPN
                      traffic
                  run_id:
                    type: integer
                    description: Algorithm's run_id
                  label:
                    type: string
                    description: Label for port specified in algorithm
                      docker image

        responses:
          201:
            description: Ok
          401:
            description: Unauthorized

        security:
            - bearerAuth: []

        tags: ["VPN"]
        """
        data = request.get_json()
        # validate request body
        errors = port_input_schema.validate(data)
        if errors:
            return {'msg': 'Request body is incorrect', 'errors': errors}, \
                HTTPStatus.BAD_REQUEST

        # The only entity that is allowed to algorithm ports is the node where
        # those algorithms are running.
        run_id = data['run_id']
        linked_run = g.session.query(Run).filter(Run.id == run_id).one()
        if g.node.id != linked_run.node.id:
            return {'msg': 'You lack the permissions to do that!'},\
                HTTPStatus.UNAUTHORIZED

        port = AlgorithmPort(
            port=data['port'],
            run_id=run_id,
            label=data.get('label' ''),
        )
        port.save()

        return port_schema.dump(port, many=False), HTTPStatus.CREATED

    @with_node
    def delete(self):
        """ Delete ports by run_id
        ---
        description: >-
          Deletes descriptions of a port that is available for VPN
          communication for a certain algorithm. The ports are deleted based
          on run_id. Only the node on which the algorithm is running is
          allowed to delete this. This happens on task completion.\n

          This endpoint is not accessible for users, but only for
          authenticated nodes.

        parameters:
          - in: path
            name: run_id
            schema:
              type: integer
            minimum: 1
            description: Run id for which ports must be deleted
            required: true

        responses:
          200:
            description: Ok
          400:
            description: Run id was not defined
          401:
            description: Unauthorized

        security:
          - bearerAuth: []

        tags: ["VPN"]
        """
        args = request.args
        if 'run_id' not in args:
            return {'msg': 'The run_id argument is required!'}, \
              HTTPStatus.BAD_REQUEST

        # The only entity that is allowed to delete algorithm ports is the node
        # where those algorithms are running.
        run_id = args['run_id']
        linked_run = g.session.query(Run).filter(Run.id == run_id).one()
        if g.node.id != linked_run.node.id:
            return {'msg': 'You lack the permissions to do that!'},\
                HTTPStatus.UNAUTHORIZED

        # all checks passed: delete the port entries
        g.session.query(AlgorithmPort).filter(
            AlgorithmPort.run_id == run_id
        ).delete()
        g.session.commit()

        return {"msg": "Ports removed from the database."}, HTTPStatus.OK


class VPNAddress(ServicesResources):

    @with_container
    def get(self):
        """
        Get a list of the addresses (IP + port) and labels of algorithm
        containers in the same task as the authenticating container.
        ---

        description: >-
          Returns a dictionary of addresses of algorithm containers in the same
          task.\n

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Port|Global|View|❌|❌|View any result|\n
          |Port|Organization|View|❌|✅|View the ports of your
          organizations collaborations|\n

          Not accessible to users.

        parameters:
          - in: path
            name: label
            schema:
              type: string
            description: Algorithm port label to filter by
          - in: path
            name: only_children
            schema:
              type: boolean
            description: Only include the addresses of subtasks, not those at
              the same level. Incompatible with 'only_parent'.
          - in: path
            name: only_parent
            schema:
              type: boolean
            description: Only send the address of the parent task, not those at
              the same level. Incompatible with 'only_children'.
          - in: path
            name: include_children
            schema:
              type: boolean
            description: Include the addresses of subtasks. Ignored if
              'only_children' is True. Incompatible with 'only_parent',
              superseded by 'only_children'.
          - in: path
            name: include_parent
            schema:
              type: boolean
            description: Include the addresses of parent tasks. Ignored if
              'only_parent' is True. Incopatible with 'only_children',
              superseded by 'only_parent'.

        responses:
          200:
            description: Ok
          400:
            description: Incompatible arguments specified

        security:
        - bearerAuth: []

        tags: ["VPN"]
        """
        task_id = g.container['task_id']
        task_ids = [task_id]

        task = db.Task.get(task_id)

        include_children = request.args.get('include_children', False)
        include_parent = request.args.get('include_parent', False)
        only_children = request.args.get('only_children', False)
        only_parent = request.args.get('only_parent', False)

        if only_children and only_parent:
            return {
                'msg': 'Using only_children and only_parent simultaneously is '
                'not possible! Specify one or the other.'
            }, HTTPStatus.BAD_REQUEST
        elif only_children and include_parent:
            return {
                'msg': 'Using only_children and include_parent simultaneously '
                'is not possible! Specify one or the other.'
            }, HTTPStatus.BAD_REQUEST
        elif only_parent and include_children:
            return {
                'msg': 'Using only_parent and include_children simultaneously '
                'is not possible! Specify one or the other.'
            }, HTTPStatus.BAD_REQUEST

        # include child tasks if requested
        if include_children or only_children:
            subtasks = g.session.query(db.Task).filter(
                db.Task.parent_id == task_id
            ).all()
            if only_children:
                task_ids = [t.id for t in subtasks]
            else:
                task_ids.extend([t.id for t in subtasks])

        # include parent task if requested
        if include_parent or only_parent:
            parent = g.session.query(db.Task).filter(
                db.Task.id == task.parent_id
            ).one_or_none()
            if parent:
                if only_parent:
                    task_ids = [parent.id]
                else:
                    task_ids.append(parent.id)

        # get all ports for the tasks requested
        q = g.session.query(AlgorithmPort)\
                     .join(Run)\
                     .filter(Run.task_id.in_(task_ids))\

        # filter by label if requested
        filter_label = request.args.get('label')
        if filter_label:
            q = q.filter(AlgorithmPort.label == filter_label)

        ports = q.all()

        # combine data from ports and nodes
        addresses = []
        for port in ports:
            d = {
                'port': port.port,
                'label': port.label,
                'ip': port.run.node.ip,
                'organization_id': port.run.organization_id,
                'task_id': port.run.task_id,
                'parent_id': port.run.task.parent_id,
            }
            addresses.append(d)

        return {'addresses': addresses}, HTTPStatus.OK



# -*- coding: utf-8 -*-
import logging

from http import HTTPStatus
from flask.globals import g
from flask_restful import Api

from vantage6.server.resource import ServicesResources
from vantage6.common import logger_name


module_name = logger_name(__name__)
log = logging.getLogger(module_name)


def setup(api: Api, api_base: str, services: dict) -> None:
    """
    Setup the health resource.

    Parameters
    ----------
    api : Api
        Flask restful api instance
    api_base : str
        Base url of the api
    services : dict
        Dictionary with services required for the resource endpoints
    """
    path = "/".join([api_base, module_name])
    log.info(f'Setting up "{path}" and subdirectories')

    api.add_resource(
        Health,
        path,
        endpoint='health',
        methods=('GET',),
        resource_class_kwargs=services
    )


# ------------------------------------------------------------------------------
# Resources / API's
# ------------------------------------------------------------------------------
class Health(ServicesResources):

    def get(self):
        """Displays the health of services
        ---
        description: >-
          Checks if server can communicate with the database. If not, it throws
          an exception.

        responses:
          200:
            description: Ok

        security:
        - bearerAuth: []

        tags: ["Database"]
        """

        # test DB
        db_ok = False
        try:
            g.session.execute('SELECT 1')
            db_ok = True
        except Exception as e:
            log.error("DB not responding")
            log.debug(e)

        return {'database': db_ok}, HTTPStatus.OK



"""
The server has a central function in the vantage6 architecture. It stores
in the database which organizations, collaborations, users, etc.
exist. It allows the users and nodes to authenticate and subsequently interact
through the API the server hosts. Finally, it also communicates with
authenticated nodes and users via the socketIO server that is run here.
"""
# -*- coding: utf-8 -*-
from gevent import monkey

# flake8: noqa: E402 (ignore import error)
monkey.patch_all()

import importlib
import logging
import uuid
import json
import time
import datetime as dt
import traceback

from http import HTTPStatus
from werkzeug.exceptions import HTTPException
from flasgger import Swagger
from flask import (
    Flask, make_response, current_app, request, send_from_directory, Request,
    Response
)
from flask_cors import CORS
from flask_jwt_extended import JWTManager
from flask_marshmallow import Marshmallow
from flask_restful import Api
from flask_mail import Mail
from flask_principal import Principal, Identity, identity_changed
from flask_socketio import SocketIO
from threading import Thread
from pathlib import Path

from vantage6.common import logger_name
from vantage6.common.globals import PING_INTERVAL_SECONDS
from vantage6.server import db
from vantage6.cli.context import ServerContext
from vantage6.server.model.base import DatabaseSessionManager, Database
from vantage6.server.resource.common.output_schema import HATEOASModelSchema
from vantage6.server.permission import RuleNeed, PermissionManager
from vantage6.server.globals import (
    APPNAME,
    ACCESS_TOKEN_EXPIRES_HOURS,
    RESOURCES,
    SUPER_USER_INFO,
    REFRESH_TOKENS_EXPIRE_HOURS,
    DEFAULT_SUPPORT_EMAIL_ADDRESS,
    MIN_TOKEN_VALIDITY_SECONDS,
    MIN_REFRESH_TOKEN_EXPIRY_DELTA,
    SERVER_MODULE_NAME
)
from vantage6.server.resource.common.swagger_templates import swagger_template
from vantage6.server._version import __version__
from vantage6.server.mail_service import MailService
from vantage6.server.websockets import DefaultSocketNamespace
from vantage6.server.default_roles import get_default_roles, DefaultRole


module_name = logger_name(__name__)
log = logging.getLogger(module_name)


class ServerApp:
    """
    Vantage6 server instance.

    Attributes
    ----------
    ctx : ServerContext
        Context object that contains the configuration of the server.
    """

    def __init__(self, ctx: ServerContext) -> None:
        """Create a vantage6-server application."""

        self.ctx = ctx

        # initialize, configure Flask
        self.app = Flask(
            SERVER_MODULE_NAME, root_path=Path(__file__),
            template_folder=Path(__file__).parent / 'templates',
            static_folder=Path(__file__).parent / 'static'
        )
        self.debug: dict = self.ctx.config.get('debug', {})
        self.configure_flask()

        # Setup SQLAlchemy and Marshmallow for marshalling/serializing
        self.ma = Marshmallow(self.app)

        # Setup the Flask-JWT-Extended extension (JWT: JSON Web Token)
        self.jwt = JWTManager(self.app)
        self.configure_jwt()

        # Setup Principal, granular API access manegement
        self.principal = Principal(self.app, use_sessions=False)

        # Enable cross-origin resource sharing
        self.cors = CORS(self.app)

        # SWAGGER documentation
        self.swagger = Swagger(self.app, template=swagger_template)

        # Setup the Flask-Mail client
        self.mail = MailService(self.app, Mail(self.app))

        # Setup websocket channel
        self.socketio = self.setup_socket_connection()

        # setup the permission manager for the API endpoints
        self.permissions = PermissionManager()

        # Api - REST JSON-rpc
        self.api = Api(self.app)
        self.configure_api()
        self.load_resources()

        # set the server version
        self.__version__ = __version__

        # set up socket ping/pong
        log.debug("Starting thread to set node status")
        t = Thread(target=self.__node_status_worker, daemon=True)
        t.start()

        log.info("Initialization done")

    def setup_socket_connection(self) -> SocketIO:
        """
        Setup a socket connection. If a message queue is defined, connect the
        socket to the message queue. Otherwise, use the default socketio
        settings.

        Returns
        -------
        SocketIO
            SocketIO object
        """

        msg_queue = self.ctx.config.get('rabbitmq', {}).get('uri')
        if msg_queue:
            log.debug(f'Connecting to msg queue: {msg_queue}')

        debug_mode = self.debug.get('socketio', False)
        if debug_mode:
            log.debug("SocketIO debug mode enabled")
        try:
            socketio = SocketIO(
                self.app,
                async_mode='gevent_uwsgi',
                message_queue=msg_queue,
                ping_timeout=60,
                cors_allowed_origins='*',
                logger=debug_mode,
                engineio_logger=debug_mode
            )
        except Exception as e:
            log.warning('Default socketio settings failed, attempt to run '
                        'without gevent_uwsgi packages! This leads to '
                        'performance issues and possible issues concerning '
                        'the websocket channels!')
            log.debug(e)
            socketio = SocketIO(
                self.app,
                message_queue=msg_queue,
                ping_timeout=60,
                cors_allowed_origins='*',
                logger=debug_mode,
                engineio_logger=debug_mode
            )

        # FIXME: temporary fix to get socket object into the namespace class
        DefaultSocketNamespace.socketio = socketio
        socketio.on_namespace(DefaultSocketNamespace("/tasks"))

        return socketio

    def configure_flask(self) -> None:
        """Configure the Flask settings of the vantage6 server."""

        # let us handle exceptions
        self.app.config['PROPAGATE_EXCEPTIONS'] = True

        # patch where to obtain token
        self.app.config['JWT_AUTH_URL_RULE'] = '/api/token'

        # If no secret is set in the config file, one is generated. This
        # implies that all (even refresh) tokens will be invalidated on restart
        self.app.config['JWT_SECRET_KEY'] = self.ctx.config.get(
            'jwt_secret_key',
            str(uuid.uuid1())
        )

        # Default expiration time
        token_expiry_seconds = self._get_jwt_expiration_seconds(
            config_key='token_expires_hours',
            default_hours=ACCESS_TOKEN_EXPIRES_HOURS
        )
        self.app.config['JWT_ACCESS_TOKEN_EXPIRES'] = token_expiry_seconds

        # Set refresh token expiration time
        self.app.config['JWT_REFRESH_TOKEN_EXPIRES'] = \
                self._get_jwt_expiration_seconds(
            config_key='refresh_token_expires_hours',
            default_hours=REFRESH_TOKENS_EXPIRE_HOURS,
            longer_than=token_expiry_seconds + MIN_REFRESH_TOKEN_EXPIRY_DELTA,
            is_refresh=True
        )

        # Open Api Specification (f.k.a. swagger)
        self.app.config['SWAGGER'] = {
            'title': APPNAME,
            'uiversion': "3",
            'openapi': '3.0.0',
            'version': __version__
        }

        # Mail settings
        mail_config = self.ctx.config.get("smtp", {})
        self.app.config["MAIL_PORT"] = mail_config.get("port", 1025)
        self.app.config["MAIL_SERVER"] = mail_config.get("server", "localhost")
        self.app.config["MAIL_USERNAME"] = mail_config.get(
            "username",
            DEFAULT_SUPPORT_EMAIL_ADDRESS
        )
        self.app.config["MAIL_PASSWORD"] = mail_config.get("password", "")
        self.app.config["MAIL_USE_TLS"] = mail_config.get("MAIL_USE_TLS",
                                                          True)
        self.app.config["MAIL_USE_SSL"] = mail_config.get("MAIL_USE_SSL",
                                                          False)
        debug_mode = self.debug.get('flask', False)
        if debug_mode:
            log.debug("Flask debug mode enabled")
        self.app.debug = debug_mode

        def _get_request_path(request: Request) -> str:
            """
            Return request extension of request URL, e.g.
            http://localhost:5000/api/task/1 -> api/task/1

            Parameters
            ----------
            request: Request
                Flask request object

            Returns
            -------
            string:
                The endpoint path of the request
            """
            return request.url.replace(request.url_root, '')

        # before request
        @self.app.before_request
        def do_before_request():
            """Before every flask request method."""
            # Add log message before each request
            log.debug(f"Received request: {request.method} "
                      f"{_get_request_path(request)}")

            # This will obtain a (scoped) db session from the session factory
            # that is linked to the flask request global `g`. In every endpoint
            # we then can access the database by using this session. We ensure
            # that the session is removed (and uncommited changes are rolled
            # back) at the end of every request.
            DatabaseSessionManager.new_session()

        @self.app.after_request
        def remove_db_session(response):
            """After every flask request.

            This will close the database session created by the
            `before_request`.
            """
            DatabaseSessionManager.clear_session()
            return response

        @self.app.errorhandler(HTTPException)
        def error_remove_db_session(error: HTTPException):
            """In case an HTTP-exception occurs during the request.

            It is important to close the db session to avoid having dangling
            sessions.
            """
            if error.code == 404:
                log.debug(
                    f"404 error for route '{_get_request_path(request)}'")
            else:
                log.warn('HTTP Exception occured during request')
                log.debug(traceback.format_exc())
            DatabaseSessionManager.clear_session()
            return error.get_response()

        @self.app.errorhandler(Exception)
        def error2_remove_db_session(error):
            """In case an exception occurs during the request.

            It is important to close the db session to avoid having dangling
            sessions.
            """
            log.exception('Exception occured during request')
            DatabaseSessionManager.clear_session()
            return {'msg': f'An unexpected error occurred on the server!'}, \
                HTTPStatus.INTERNAL_SERVER_ERROR

        @self.app.route('/robots.txt')
        def static_from_root():
            return send_from_directory(self.app.static_folder,
                                       request.path[1:])

    def _get_jwt_expiration_seconds(
        self, config_key: str, default_hours: int,
        longer_than: int = MIN_TOKEN_VALIDITY_SECONDS,
        is_refresh: bool = False
    ) -> int:
        """
        Return the expiration time for JWT tokens.

        This time may be specified in the config file. If it is not, the
        default value is returned.

        Parameters
        ----------
        config_key: str
            The config key to look for that sets the expiration time
        default_hours: int
            The default expiration time in hours
        longer_than: int
            The minimum expiration time in hours.
        is_refresh: bool
            If True, the expiration time is for a refresh token. If False, it
            is for an access token.

        Returns
        -------
        int:
            The JWT token expiration time in seconds
        """
        hours_expire = self.ctx.config.get(config_key)
        if hours_expire is None:
            # No value is present in the config file, use default
            seconds_expire = int(float(default_hours) * 3600)
        elif isinstance(hours_expire, (int, float)) or \
                hours_expire.replace(".", "").isnumeric():
            # Numeric value is present in the config file
            seconds_expire = int(float(hours_expire) * 3600)
            if seconds_expire < longer_than:
                log.warning(
                    f"Invalid value for '{config_key}': {hours_expire}. Tokens"
                    f" must be valid for at least {longer_than} seconds. Using"
                    f" default value: {default_hours} hours")
                if is_refresh:
                    log.warning("Note that refresh tokens should be valid at "
                                f"least {MIN_REFRESH_TOKEN_EXPIRY_DELTA} "
                                "seconds longer than access tokens.")
                seconds_expire = int(float(default_hours) * 3600)
        else:
            # Non-numeric value is present in the config file. Warn and use
            # default
            log.error(f"Invalid value for '{config_key}': {hours_expire}. "
                      f"Using default value: {default_hours} hours")
            seconds_expire = int(float(default_hours) * 3600)

        return seconds_expire

    def configure_api(self) -> None:
        """Define global API output and its structure."""

        # helper to create HATEOAS schemas
        HATEOASModelSchema.api = self.api

        # whatever you get try to json it
        @self.api.representation('application/json')
        # pylint: disable=unused-argument
        def output_json(
            data: db.Base | list[db.Base], code: HTTPStatus,
            headers: dict = None
        ) -> Response:
            """
            Return jsonified data for request responses.

            Parameters
            ----------
            data: db.Base | list[db.Base]
                The data to be jsonified
            code: HTTPStatus
                The HTTP status code of the response
            headers: dict
                Additional headers to be added to the response
            """

            if isinstance(data, db.Base):
                data = db.jsonable(data)
            elif isinstance(data, list) and len(data) and \
                    isinstance(data[0], db.Base):
                data = db.jsonable(data)

            resp = make_response(json.dumps(data), code)
            resp.headers.extend(headers or {})
            return resp

    def configure_jwt(self):
        """Configure JWT authentication."""

        @self.jwt.additional_claims_loader
        # pylint: disable=unused-argument
        def additional_claims_loader(
                identity: db.Authenticatable | dict) -> dict:
            """
            Create additional claims for JWT tokens: set user type and for
            users, set their roles.

            Parameters
            ----------
            identity: db.Authenticatable | dict
                The identity for which to create the claims

            Returns
            -------
            dict:
                The claims to be added to the JWT token
            """
            roles = []
            if isinstance(identity, db.User):
                type_ = 'user'
                roles = [role.name for role in identity.roles]

            elif isinstance(identity, db.Node):
                type_ = 'node'
            elif isinstance(identity, dict):
                type_ = 'container'
            else:
                log.error(f"could not create claims from {str(identity)}")
                return

            claims = {
                'client_type': type_,
                'roles': roles,
            }

            return claims

        @self.jwt.user_identity_loader
        # pylint: disable=unused-argument
        def user_identity_loader(
                identity: db.Authenticatable | dict) -> str | dict:
            """"
            JSON serializing identity to be used by ``create_access_token``.

            Parameters
            ----------
            identity: db.Authenticatable | dict
                The identity to be serialized

            Returns
            -------
            str | dict:
                The serialized identity. For a node or user, this is the id;
                for a container, it is a dict.
            """
            if isinstance(identity, db.Authenticatable):
                return identity.id
            if isinstance(identity, dict):
                return identity

            log.error(f"Could not create a JSON serializable identity \
                        from '{str(identity)}'")

        @self.jwt.user_lookup_loader
        # pylint: disable=unused-argument
        def user_lookup_loader(
            jwt_payload: dict, jwt_headers: dict
        ) -> db.Authenticatable | dict:
            """
            Load the user, node or container instance from the JWT payload.

            Parameters
            ----------
            jwt_payload: dict
                The JWT payload
            jwt_headers: dict
                The JWT headers

            Returns
            -------
            db.Authenticatable | dict:
                The user, node or container identity. If the identity is a
                container, a dict is returned.
            """
            identity = jwt_headers['sub']
            auth_identity = Identity(identity)

            # in case of a user or node an auth id is shared as identity
            if isinstance(identity, int):

                # auth_identity = Identity(identity)

                auth = db.Authenticatable.get(identity)

                if isinstance(auth, db.Node):

                    for rule in db.Role.get_by_name(DefaultRole.NODE).rules:
                        auth_identity.provides.add(
                                RuleNeed(
                                    name=rule.name,
                                    scope=rule.scope,
                                    operation=rule.operation
                                )
                            )

                if isinstance(auth, db.User):

                    # add role permissions
                    for role in auth.roles:
                        for rule in role.rules:
                            auth_identity.provides.add(
                                RuleNeed(
                                    name=rule.name,
                                    scope=rule.scope,
                                    operation=rule.operation
                                )
                            )

                    # add 'extra' permissions
                    for rule in auth.rules:
                        auth_identity.provides.add(
                            RuleNeed(
                                name=rule.name,
                                scope=rule.scope,
                                operation=rule.operation
                            )
                        )

                identity_changed.send(current_app._get_current_object(),
                                      identity=auth_identity)

                return auth
            else:
                # container identity

                for rule in db.Role.get_by_name(DefaultRole.CONTAINER).rules:
                    auth_identity.provides.add(
                        RuleNeed(
                            name=rule.name,
                            scope=rule.scope,
                            operation=rule.operation
                        )
                    )
                identity_changed.send(current_app._get_current_object(),
                                      identity=auth_identity)
                log.debug(identity)
                return identity

    def load_resources(self) -> None:
        """Import the modules containing API resources."""

        # make services available to the endpoints, this way each endpoint can
        # make use of 'em.
        services = {
            "socketio": self.socketio,
            "mail": self.mail,
            "api": self.api,
            "permissions": self.permissions,
            "config": self.ctx.config
        }

        for res in RESOURCES:
            module = importlib.import_module('vantage6.server.resource.' + res)
            module.setup(self.api, self.ctx.config['api_path'], services)

    # TODO consider moving this method elsewhere. This is not trivial at the
    # moment because of the circular import issue with `db`, see
    # https://github.com/vantage6/vantage6/issues/53
    @staticmethod
    def _add_default_roles() -> None:
        for role in get_default_roles(db):
            if not db.Role.get_by_name(role['name']):
                log.warn(f"Creating new default role {role['name']}...")
                new_role = db.Role(
                    name=role['name'],
                    description=role['description'],
                    rules=role['rules']
                )
                new_role.save()

    def start(self) -> None:
        """
        Start the server.

        Before server is really started, some database settings are checked and
        (re)set where appropriate.
        """
        # add default roles (if they don't exist yet)
        self._add_default_roles()

        # create root user if it is not in the DB yet
        try:
            db.User.get_by_username(SUPER_USER_INFO['username'])
        except Exception:
            log.warn("No root user found! Is this the first run?")

            log.debug("Creating organization for root user")
            org = db.Organization(name="root")

            # TODO use constant instead of 'Root' literal
            root = db.Role.get_by_name("Root")

            log.warn(f"Creating root user: "
                     f"username={SUPER_USER_INFO['username']}, "
                     f"password={SUPER_USER_INFO['password']}")

            user = db.User(username=SUPER_USER_INFO['username'], roles=[root],
                           organization=org, email="root@domain.ext",
                           password=SUPER_USER_INFO['password'],
                           failed_login_attempts=0,
                           last_login_attempt=None)
            user.save()
        return self

    def __node_status_worker(self) -> None:
        """
        Set node status to offline if they haven't send a ping message in a
        while.
        """
        # start periodic check if nodes are responsive
        while True:
            # Send ping event
            try:
                before_wait = dt.datetime.utcnow()

                # Wait a while to give nodes opportunity to pong. This interval
                # is a bit longer than the interval at which the nodes ping,
                # because we want to make sure that the nodes have had time to
                # respond.
                time.sleep(PING_INTERVAL_SECONDS + 5)

                # Check for each node that is online if they have responded.
                # Otherwise set them to offline.
                online_status_nodes = db.Node.get_online_nodes()
                for node in online_status_nodes:
                    if node.last_seen < before_wait:
                        node.status = 'offline'
                        node.save()
            except Exception:
                log.exception('Node-status thread had an exception')
                time.sleep(PING_INTERVAL_SECONDS)


def run_server(config: str, system_folders: bool = True) -> ServerApp:
    """
    Run a vantage6 server.

    Parameters
    ----------
    config: str
        Configuration file path
    system_folders: bool
        Whether to use system or user folders. Default is True.

    Returns
    -------
    ServerApp
        A running instance of the vantage6 server
    """
    ctx = ServerContext.from_external_config_file(
        config,
        system_folders
    )
    allow_drop_all = ctx.config["allow_drop_all"]
    Database().connect(uri=ctx.get_database_uri(),
                       allow_drop_all=allow_drop_all)
    return ServerApp(ctx).start()


def run_dev_server(server_app: ServerApp, *args, **kwargs) -> None:
    """
    Run a vantage6 development server (outside of a Docker container).

    Parameters
    ----------
    server_app: ServerApp
        Instance of a vantage6 server
    """
    log.warn('*'*80)
    log.warn(' DEVELOPMENT SERVER '.center(80, '*'))
    log.warn('*'*80)
    kwargs.setdefault('log_output', False)
    server_app.socketio.run(server_app.app, *args, **kwargs)



from __future__ import annotations
import base64

from sqlalchemy import Column, String, LargeBinary
from sqlalchemy.orm import relationship
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm.exc import NoResultFound

from vantage6.common.globals import STRING_ENCODING
from vantage6.server.model.base import Base, DatabaseSessionManager


class Organization(Base):
    """Table that describes which organizations are available.

    An organization is the legal entity that plays a central role in managing
    distributed tasks. Each organization contains a public key which other
    organizations can use to send encrypted messages that only this
    organization can read.

    Attributes
    ----------
    name : str
        Name of the organization
    domain : str
        Domain of the organization
    address1 : str
        Address of the organization
    address2 : str
        Address of the organization
    zipcode : str
        Zipcode of the organization
    country : str
        Country of the organization
    _public_key : bytes
        Public key of the organization
    collaborations :
            list[:class:`~vantage6.server.model.collaboration.Collaboration`]
        List of collaborations that this organization is part of
    results : list[:class:`~vantage6.server.model.run.Run`]
        List of results that are part of this organization
    nodes : list[:class:`~vantage6.server.model.node.Node`]
        List of nodes that are part of this organization
    users : list[:class:`~vantage6.server.model.user.User`]
        List of users that are part of this organization
    tasks : list[:class:`~vantage6.server.model.task.Task`]
        List of tasks that are created by this organization
    roles : list[:class:`~vantage6.server.model.role.Role`]

    """
    # fields
    name = Column(String, unique=True)
    domain = Column(String)
    address1 = Column(String)
    address2 = Column(String)
    zipcode = Column(String)
    country = Column(String)
    _public_key = Column(LargeBinary)

    # relations
    collaborations = relationship("Collaboration", secondary="Member",
                                  back_populates="organizations")
    runs = relationship("Run", back_populates="organization")
    nodes = relationship("Node", back_populates="organization")
    users = relationship("User", back_populates="organization")
    tasks = relationship("Task", back_populates="init_org")
    roles = relationship("Role", back_populates="organization")

    @classmethod
    def get_by_name(cls, name) -> Organization | None:
        """
        Returns the organization with the given name.

        Parameters
        ----------
        name : str
            Name of the organization

        Returns
        -------
        Organization | None
            Organization with the given name if it exists, otherwise None
        """
        session = DatabaseSessionManager.get_session()
        try:
            result = session.query(cls).filter_by(name=name).first()
            session.commit()
            return result
        except NoResultFound:
            return None

    @hybrid_property
    def public_key(self) -> str:
        """
        Returns the public key of the organization.

        Returns
        -------
        str
            Public key of the organization. Empty string if no public key is
            set.
        """
        if self._public_key:
            # TODO this should be fixed properly
            try:
                return base64.b64decode(self._public_key)\
                    .decode(STRING_ENCODING)
            except Exception:
                return ""
        else:
            return ""

    @public_key.setter
    def public_key(self, public_key_b64: str) -> None:
        """
        Set public key of the organization. Assumes that the public key is
        already b64-encoded.

        Parameters
        ----------
        public_key_b64 : str
            Public key of the organization, b64-encoded
        """
        self._public_key = base64.b64decode(
            public_key_b64.encode(STRING_ENCODING)
        )

    def __repr__(self) -> str:
        """
        Returns a string representation of the organization.

        Returns
        -------
        str
            String representation of the organization
        """
        number_of_users = len(self.users)
        return (
            "<Organization "
            f"{self.id}: '{self.name}', "
            f"domain:{self.domain}, "
            f"users:{number_of_users}"
            ">"
        )



# -*- coding: utf-8 -*-
import logging

from flask import request, g
from flask_restful import Api
from http import HTTPStatus

from vantage6.common import logger_name
from vantage6.server import db
from vantage6.server.resource.common.pagination import Pagination
from vantage6.server.permission import (
    Scope as S,
    Operation as P,
    PermissionManager,
    RuleCollection
)
from vantage6.server.resource.common.input_schema import (
    OrganizationInputSchema
)
from vantage6.server.resource import (
    only_for, with_user, ServicesResources
)
from vantage6.server.resource.common.output_schema import OrganizationSchema


module_name = logger_name(__name__)
log = logging.getLogger(module_name)


def setup(api: Api, api_base: str, services: dict) -> None:
    """
    Setup the organization resource.

    Parameters
    ----------
    api : Api
        Flask restful api instance
    api_base : str
        Base url of the api
    services : dict
        Dictionary with services required for the resource endpoints
    """
    path = "/".join([api_base, module_name])
    log.info(f'Setting up "{path}" and subdirectories')

    api.add_resource(
        Organizations,
        path,
        endpoint='organization_without_id',
        methods=('GET', 'POST'),
        resource_class_kwargs=services
     )
    api.add_resource(
        Organization,
        path + '/<int:id>',
        endpoint='organization_with_id',
        methods=('GET', 'PATCH'),
        resource_class_kwargs=services
    )


# -----------------------------------------------------------------------------
# Permissions
# -----------------------------------------------------------------------------
def permissions(permissions: PermissionManager) -> None:
    """
    Define the permissions for this resource.

    Parameters
    ----------
    permissions : PermissionManager
        Permission manager instance to which permissions are added
    """
    add = permissions.appender(module_name)

    add(scope=S.GLOBAL, operation=P.VIEW,
        description="view any organization")
    add(scope=S.ORGANIZATION, operation=P.VIEW,
        description="view your own organization info",
        assign_to_container=True, assign_to_node=True)
    add(scope=S.COLLABORATION, operation=P.VIEW,
        description='view collaborating organizations',
        assign_to_container=True, assign_to_node=True)
    add(scope=S.GLOBAL, operation=P.EDIT,
        description="edit any organization")
    add(scope=S.ORGANIZATION, operation=P.EDIT,
        description="edit your own organization info", assign_to_node=True)
    add(scope=S.COLLABORATION, operation=P.EDIT,
        description='edit collaborating organizations')
    add(scope=S.GLOBAL, operation=P.CREATE,
        description="create a new organization")


# ------------------------------------------------------------------------------
# Resources / API's
# ------------------------------------------------------------------------------
org_schema = OrganizationSchema()
org_input_schema = OrganizationInputSchema()


class OrganizationBase(ServicesResources):

    def __init__(self, socketio, mail, api, permissions, config):
        super().__init__(socketio, mail, api, permissions, config)
        self.r: RuleCollection = getattr(self.permissions, module_name)


class Organizations(OrganizationBase):

    @only_for(("user", "node", "container"))
    def get(self):
        """ Returns a list organizations
        ---
        description: >-
            Get a list of organizations based on filters and user permissions\n

            ### Permission Table\n
            |Rule name|Scope|Operation|Assigned to node|Assigned to container|
            Description|\n
            |--|--|--|--|--|--|\n
            |Organization|Global|View|❌|❌|View all organizations|\n
            |Organization|Collaboration|View|✅|✅|View a list of organizations
            within the scope of the collaboration|\n
            |Organization|Organization|View|✅|✅|View a 'list' of just the
            organization you are part of|\n

            Accessible to users.

        parameters:
          - in: query
            name: name
            schema:
              type: string
            description: >-
              Name to match with a LIKE operator. \n
              * The percent sign (%) represents zero, one, or multiple
              characters\n
              * underscore sign (_) represents one, single character
          - in: query
            name: country
            schema:
              type: string
            description: Country
          - in: query
            name: collaboration_id
            schema:
              type: integer
            description: Collaboration id
          - in: query
            name: page
            schema:
              type: integer
            description: Page number for pagination (default=1)
          - in: query
            name: per_page
            schema:
              type: integer
            description: Number of items per page (default=10)
          - in: query
            name: sort
            schema:
              type: string
            description: >-
              Sort by one or more fields, separated by a comma. Use a minus
              sign (-) in front of the field to sort in descending order.

        responses:
          200:
            description: Ok
          401:
            description: Unauthorized
          400:
            description: Improper values for pagination or sorting parameters

        security:
          - bearerAuth: []

        tags: ["Organization"]
        """

        # Obtain the organization of the requester
        auth_org = self.obtain_auth_organization()
        args = request.args

        # query
        q = g.session.query(db.Organization)
        g.session.commit()

        # filter by a field of this endpoint
        if 'name' in args:
            q = q.filter(db.Organization.name.like(args['name']))
        if 'country' in args:
            q = q.filter(db.Organization.country == args['country'])
        if 'collaboration_id' in args:
            if not self.r.can_for_col(P.VIEW, int(args['collaboration_id'])):
                return {
                    'msg': 'You lack the permission to get all organizations '
                    'in your collaboration!'
                }, HTTPStatus.UNAUTHORIZED
            q = q.join(db.Member).join(db.Collaboration)\
                 .filter(db.Collaboration.id == args['collaboration_id'])

        # filter the list of organizations based on the scope
        if self.r.v_glo.can():
            pass  # don't apply filters
        elif self.r.v_col.can():
            # obtain collaborations your organization participates in
            collabs = g.session.query(db.Collaboration).filter(
                db.Collaboration.organizations.any(id=auth_org.id)
            ).all()
            g.session.commit()

            # filter orgs in own collaborations, and add own organization in
            # case this organization does not participate in any collaborations
            # yet
            org_ids = [o.id for col in collabs for o in col.organizations]
            org_ids = list(set(org_ids + [auth_org.id]))

            # select only the organizations in the collaborations
            q = q.filter(db.Organization.id.in_(org_ids))

        elif self.r.v_org.can():
            q = q.filter(db.Organization.id == auth_org.id)
        else:
            return {'msg': 'You lack the permission to do that!'}, \
                HTTPStatus.UNAUTHORIZED

        # paginate the results
        try:
            page = Pagination.from_query(query=q, request=request)
        except ValueError as e:
            return {'msg': str(e)}, HTTPStatus.BAD_REQUEST

        # serialization of DB model
        return self.response(page, org_schema)

    @with_user
    def post(self):
        """Create new organization
        ---
        description: >-
          Creates a new organization from the specified values\n

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Organization|Global|Create|❌|❌|Create a new organization|\n

          Accessible to users.

        requestBody:
          content:
            application/json:
              schema:
                $ref: '#/components/schemas/Organization'

        responses:
          201:
            description: Ok
          401:
            description: Unauthorized
          400:
            description: Organization with that name already exists

        security:
          - bearerAuth: []

        tags: ["Organization"]
        """

        if not self.r.c_glo.can():
            return {'msg': 'You lack the permissions to do that!'},\
                HTTPStatus.UNAUTHORIZED

        # validate request body
        data = request.get_json()
        errors = org_input_schema.validate(data)
        if errors:
            return {'msg': 'Request body is incorrect', 'errors': errors}, \
                HTTPStatus.BAD_REQUEST

        name = data.get('name')
        if db.Organization.exists("name", name):
            return {
                "msg": f"Organization with name '{name}' already exists!"
            }, HTTPStatus.BAD_REQUEST

        organization = db.Organization(
            name=name,
            address1=data.get('address1', ''),
            address2=data.get('address2' ''),
            zipcode=data.get('zipcode', ''),
            country=data.get('country', ''),
            public_key=data.get('public_key', ''),
            domain=data.get('domain', '')
        )
        organization.save()

        return org_schema.dump(organization, many=False), \
            HTTPStatus.CREATED


class Organization(OrganizationBase):

    @only_for(("user", "node", "container"))
    def get(self, id):
        """Get organization
        ---
        description: >-
          Returns the organization specified by the id\n

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Organization|Global|View|❌|❌|View all organizations|\n
          |Organization|Collaboration|View|✅|✅|View a list of organizations
          within the scope of the collaboration|\n
          |Organization|Organization|View|✅|✅|View a list of organizations
          that the user is part of|\n

          Accessible to users.

        parameters:
          - in: path
            name: id
            schema:
              type: integer
            description: Organization id
            required: true

        responses:
          200:
            description: Ok
          404:
            description: Organization not found
          401:
            description: Unauthorized

        security:
          - bearerAuth: []

        tags: ["Organization"]
        """

        # retrieve requested organization
        req_org = db.Organization.get(id)
        if not req_org:
            return {'msg': f'Organization id={id} not found!'}, \
                HTTPStatus.NOT_FOUND

        # Check if auth has enough permissions
        if not self.r.can_for_org(P.VIEW, id):
            return {'msg': 'You do not have permission to do that!'}, \
                HTTPStatus.UNAUTHORIZED

        return org_schema.dump(req_org, many=False), HTTPStatus.OK

    @only_for(("user", "node"))
    def patch(self, id):
        """Update organization
        ---
        description: >-
          Updates the organization with the specified id.\n

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Organization|Global|Edit|❌|❌|Update an organization with
          specified id|\n
          |Organization|Collaboration|Edit|❌|❌|Update an organization within
          the collaboration the user is part of|\n
          |Organization|Organization|Edit|❌|❌|Update the organization that
          the user is part of|\n

          Accessible to users.

        parameters:
          - in: path
            name: id
            schema:
              type: integer
            description: Organization id
            required: tr

        requestBody:
          content:
            application/json:
              schema:
                $ref: '#/components/schemas/Organization'

        responses:
          200:
            description: Ok
          404:
            description: Organization with specified id is not found
          401:
            description: Unauthorized
          400:
            description: Organization with that name already exists

        security:
          - bearerAuth: []

        tags: ["Organization"]
        """
        # validate request body
        data = request.get_json()
        errors = org_input_schema.validate(data, partial=True)
        if errors:
            return {'msg': 'Request body is incorrect', 'errors': errors}, \
                HTTPStatus.BAD_REQUEST

        organization = db.Organization.get(id)
        if not organization:
            return {"msg": f"Organization with id={id} not found"}, \
                HTTPStatus.NOT_FOUND

        if not self.r.can_for_org(P.EDIT, id):
            return {'msg': 'You lack the permission to do that!'}, \
                HTTPStatus.UNAUTHORIZED

        name = data.get('name', None)
        if name:
            if organization.name != name and \
                    db.Organization.exists("name", name):
                return {
                    "msg": f"Organization with name '{name}' already exists!"
                }, HTTPStatus.BAD_REQUEST
            organization.name = name

        fields = ["address1", "address2", "zipcode", "country",
                  "public_key", "domain"]
        for field in fields:
            if field in data and data[field] is not None:
                setattr(organization, field, data[field])

        organization.save()
        return org_schema.dump(organization, many=False), HTTPStatus.OK



from __future__ import annotations
import bcrypt

from vantage6.server.model.base import DatabaseSessionManager
from sqlalchemy.orm import relationship, validates
from sqlalchemy import Column, Integer, String, ForeignKey

from vantage6.server.model.authenticatable import Authenticatable


class Node(Authenticatable):
    """
    Table that contains all registered nodes.

    Attributes
    ----------
    id : int
        Primary key
    name : str
        Name of the node
    api_key : str
        API key of the node
    collaboration : :class:`~.model.collaboration.Collaboration`
        Collaboration that the node belongs to
    organization : :class:`~.model.organization.Organization`
        Organization that the node belongs to
    """
    _hidden_attributes = ['api_key']

    id = Column(Integer, ForeignKey('authenticatable.id'), primary_key=True)

    # fields
    name = Column(String, unique=True)
    api_key = Column(String)
    collaboration_id = Column(Integer, ForeignKey("collaboration.id"))
    organization_id = Column(Integer, ForeignKey("organization.id"))

    # relationships
    collaboration = relationship("Collaboration", back_populates='nodes')
    organization = relationship("Organization", back_populates='nodes')
    config = relationship("NodeConfig", back_populates="node")

    # the type specification in Authenticatable
    __mapper_args__ = {
        'polymorphic_identity': 'node',
    }

    @validates("api_key")
    def _validate_api_key(self, key: str, api_key: str) -> str:
        """
        Hashes the api_key before storing it in the database.

        Parameters
        ----------
        key : str
            The name of the attribute that is being validated
        api_key : str
            The value of the attribute that is being validated

        Returns
        -------
        str
            The hashed api_key
        """
        return self.hash(api_key)

    def check_key(self, key: str) -> bool:
        """
        Checks if the provided key matches the stored key.

        Parameters
        ----------
        key : str
            The key to check

        Returns
        -------
        bool
            True if the provided key matches the stored key, False otherwise
        """
        if self.api_key is not None:
            expected_hash = self.api_key.encode('utf8')
            return bcrypt.checkpw(key.encode('utf8'), expected_hash)
        return False

    @classmethod
    def get_by_api_key(cls, api_key: str) -> Node | None:
        """
        Returns Node based on the provided API key.

        Parameters
        ----------
        api_key : str
            The API key of the node to search for

        Returns
        -------
        Node | None
            Returns the node if a node is associated with api_key, None if no
            node is associated with api_key.
        """
        session = DatabaseSessionManager.get_session()

        nodes = session.query(cls).all()
        session.commit()
        for node in nodes:
            is_correct_key = node.check_key(api_key)
            if is_correct_key:
                return node
        # no node found with matching API key
        return None

    @classmethod
    def get_online_nodes(cls) -> list[Node]:
        """
        Return nodes that currently have status 'online'

        Returns
        -------
        list[Node]
            List of node models that are currently online
        """
        session = DatabaseSessionManager.get_session()

        result = session.query(cls).filter_by(status='online').all()
        session.commit()
        return result

    @classmethod
    def exists_by_id(cls, organization_id: int, collaboration_id: int) -> bool:
        """
        Check if a node exists for the given organization and collaboration.

        Parameters
        ----------
        organization_id : int
            The id of the organization
        collaboration_id : int
            The id of the collaboration

        Returns
        -------
        bool
            True if a node exists for the given organization and collaboration,
            False otherwise.
        """
        session = DatabaseSessionManager.get_session()
        result = session.query(cls).filter_by(
            organization_id=organization_id,
            collaboration_id=collaboration_id
        ).scalar()
        session.commit()
        return result

    def __repr__(self) -> str:
        """
        String representation of the Node model.

        Returns
        -------
        str
            String representation of the Node model
        """
        return (
            "<Node "
            f"{self.id}: '{self.name}', "
            f"organization: {self.organization.name}, "
            f"collaboration: {self.collaboration.name}, "
            f"api_key: {self.api_key} "
            ">"
        )



import logging
import jwt
import datetime as dt

from flask import request, session
from flask_jwt_extended import get_jwt_identity, verify_jwt_in_request
from flask_socketio import Namespace, emit, join_room, leave_room

from vantage6.common import logger_name
from vantage6.common.task_status import has_task_failed
from vantage6.server import db
from vantage6.server.model.authenticatable import Authenticatable
from vantage6.server.model.rule import Operation, Scope
from vantage6.server.model.base import DatabaseSessionManager

ALL_NODES_ROOM = 'all_nodes'


class DefaultSocketNamespace(Namespace):
    """
    This is the default SocketIO namespace. It is used for all the long-running
    socket communication between the server and the clients. The clients of the
    socket connection are nodes and users.

    When socket communication is received from one of the clients, the
    functions in this class are called to execute the corresponding action.
    """
    socketio = None

    log = logging.getLogger(logger_name(__name__))

    def on_connect(self) -> None:
        """
        A new incoming connection request from a client.

        New connections are authenticated using their JWT authorization token
        which is obtained from the REST API. A session is created for each
        connected client, and lives as long as the connection is active.
        Each client is assigned to rooms based on their permissions.

        Nodes that are connecting are also set to status 'online'.


        Note
        ----
        Note that reconnecting clients are treated the same as new clients.

        """
        self.log.info(f'Client connected: "{request.sid}"')

        # try to catch jwt authorization token.
        try:
            verify_jwt_in_request()

        except jwt.exceptions.ExpiredSignatureError:
            self.log.error("JWT has expired")
            emit("expired_token", room=request.sid)
            return

        except Exception as e:
            self.log.error("Couldn't connect client! No or Invalid JWT token?")
            self.log.exception(e)
            session.name = "not-sure-yet"
            self.__join_room_and_notify(request.sid)

            # FIXME: expired probably doesn't cover it ...
            emit("expired_token", room=request.sid)
            return

        # get identity from token.
        session.auth_id = get_jwt_identity()
        auth = db.Authenticatable.get(session.auth_id)
        auth.status = 'online'
        auth.save()

        # It appears to be necessary to use the root socketio instance
        # otherwise events cannot be sent outside the current namespace.
        # In this case, only events to '/tasks' can be emitted otherwise.
        if auth.type == 'node':
            self.socketio.emit('node-status-changed', namespace='/admin')

        # define socket-session variables.
        session.type = auth.type
        session.name = auth.username if session.type == 'user' else auth.name
        self.log.info(
            f'Client identified as <{session.type}>: <{session.name}>'
        )

        # join appropiate rooms
        session.rooms = []
        if session.type == 'node':
            self._add_node_to_rooms(auth)
            self.__alert_node_status(online=True, node=auth)
        elif session.type == 'user':
            self._add_user_to_rooms(auth)

        for room in session.rooms:
            self.__join_room_and_notify(room)

        # cleanup (e.g. database session)
        self.__cleanup()

    @staticmethod
    def _add_node_to_rooms(node: Authenticatable) -> None:
        """
        Connect node to appropriate websocket rooms

        Parameters
        ----------
        node: Authenticatable
            Node that is to be added to the rooms
        """
        # node join rooms for all nodes and rooms for their collaboration
        session.rooms.append(ALL_NODES_ROOM)
        session.rooms.append(f'collaboration_{node.collaboration_id}')
        session.rooms.append(
            f'collaboration_{node.collaboration_id}_organization_'
            f'{node.organization_id}')

    @staticmethod
    def _add_user_to_rooms(user: Authenticatable) -> None:
        """
        Connect user to appropriate websocket rooms

        Parameters
        ----------
        user: Authenticatable
            User that is to be added to the rooms
        """
        # check for which collab rooms the user has permission to enter
        session.user = db.User.get(session.auth_id)
        if session.user.can('event', Scope.GLOBAL, Operation.RECEIVE):
            # user joins all collaboration rooms
            collabs = db.Collaboration.get()
            for collab in collabs:
                session.rooms.append(f'collaboration_{collab.id}')
        elif session.user.can(
                'event', Scope.COLLABORATION, Operation.RECEIVE):
            # user joins all collaboration rooms that their organization
            # participates in
            for collab in user.organization.collaborations:
                session.rooms.append(f'collaboration_{collab.id}')
        elif session.user.can('event', Scope.ORGANIZATION, Operation.RECEIVE):
            # user joins collaboration subrooms that include only messages
            # relevant to their own node
            for collab in user.organization.collaborations:
                session.rooms.append(
                    f'collaboration_{collab.id}_organization_'
                    f'{user.organization.id}'
                )

    def on_disconnect(self) -> None:
        """
        Client that disconnects is removed from all rooms they were in.

        If nodes disconnect, their status is also set to offline and users may
        be alerted to that. Also, any information on the node (e.g.
        configuration) is removed from the database.
        """
        if not self.__is_identified_client():
            self.log.debug('Client disconnected before identification')
            return

        for room in session.rooms:
            # self.__leave_room_and_notify(room)
            self.__leave_room_and_notify(room)

        auth = db.Authenticatable.get(session.auth_id)
        auth.status = 'offline'
        auth.save()

        # It appears to be necessary to use the root socketio instance
        # otherwise events cannot be sent outside the current namespace.
        # In this case, only events to '/tasks' can be emitted otherwise.
        if session.type == 'node':
            self.log.warning('emitting to /admin')
            self.socketio.emit('node-status-changed', namespace='/admin')
            self.__alert_node_status(online=False, node=auth)

            # delete any data on the node stored on the server (e.g.
            # configuration data)
            self.__clean_node_data(auth)

        self.log.info(f'{session.name} disconnected')

        # cleanup (e.g. database session)
        self.__cleanup()

    def on_message(self, message: str) -> None:
        """
        On receiving a message from a client, log it.

        Parameters
        ----------
        message: str
            Message that is going to be displayed in the server log
        """
        self.log.info('received message: ' + message)

    def on_error(self, e: str) -> None:
        """
        An receiving an error from a client, log it.

        Parameters
        ----------
        e: str
            Error message that is being displayed in the server log
        """
        self.log.error(e)

    def on_algorithm_status_change(self, data: dict) -> None:
        """
        An algorithm container has changed its status. This status change may
        be that the algorithm has finished, crashed, etc. Here we notify the
        collaboration of the change.

        Parameters
        ----------
        data: Dict
            Dictionary containing parameters on the updated algorithm status.
            It should look as follows:

            .. code:: python

                {
                    # node_id where algorithm container was running
                    "node_id": 1,
                    # new status of algorithm container
                    "status": "active",
                    # result_id for which the algorithm was running
                    "result_id": 1,
                    # collaboration_id for which the algorithm was running
                    "collaboration_id": 1
                }
        """
        # only allow nodes to send this event
        if session.type != 'node':
            self.log.warn('Only nodes can send algorithm status changes! '
                          f'{session.type} {session.auth_id} is not allowed.')
            return

        run_id = data.get('run_id')
        task_id = data.get('task_id')
        collaboration_id = data.get('collaboration_id')
        status = data.get('status')
        node_id = data.get('node_id')
        organization_id = data.get('organization_id')
        parent_id = data.get('parent_id')

        job_id = db.Run.get(run_id).task.job_id

        # log event in server logs
        msg = (f"A container for job_id={job_id} and run_id={run_id} "
               f"in collaboration_id={collaboration_id} on node_id={node_id}")
        if has_task_failed(status):
            self.log.critical(f"{msg} exited with status={status}.")
        else:
            self.log.info(f"{msg} has a new status={status}.")

        # emit task status change to other nodes/users in the collaboration
        emit(
            "algorithm_status_change", {
                "status": status,
                "run_id": run_id,
                "task_id": task_id,
                "job_id": job_id,
                "collaboration_id": collaboration_id,
                "node_id": node_id,
                "organization_id": organization_id,
                "parent_id": parent_id,
            }, room=f"collaboration_{collaboration_id}"
        )

        # cleanup (e.g. database session)
        self.__cleanup()

    def on_node_info_update(self, node_config: dict) -> None:
        """
        A node sends information about its configuration and other properties.
        Store this in the database for the duration of the node's session.

        Parameters
        ----------
        node_config: dict
            Dictionary containing the node's configuration.
        """
        # only allow nodes to send this event
        if session.type != 'node':
            self.log.warn('Only nodes can send node configuration updates! '
                          f'{session.type} {session.auth_id} is not allowed.')
            return

        node = db.Node.get(session.auth_id)

        # delete any old data that may be present (if cleanup on disconnect
        # failed)
        self.__clean_node_data(node=node)

        # store (new) node config
        to_store = []
        for k, v in node_config.items():
            # add single item or list of items
            if isinstance(v, list):
                to_store.extend([
                    db.NodeConfig(node_id=node.id, key=k, value=i)
                    for i in v
                ])
            elif isinstance(v, dict):
                to_store.extend([
                    db.NodeConfig(node_id=node.id, key=inner_key,
                                  value=inner_val)
                    for inner_key, inner_val in v.items()
                ])
            else:
                to_store.append(db.NodeConfig(node_id=node.id, key=k, value=v))

        node.config = to_store
        node.save()

        # cleanup (e.g. database session)
        self.__cleanup()

    def on_ping(self) -> None:
        """
        A client sends a ping to the server. The server detects who sent the
        ping and sets them as online.
        """
        auth = db.Authenticatable.get(session.auth_id)
        auth.status = 'online'
        auth.last_seen = dt.datetime.utcnow()
        auth.save()

    def __join_room_and_notify(self, room: str) -> None:
        """
        Joins room and notify other clients in this room.

        Parameters
        ----------
        room : str
            name of the room the client want to join
        """
        join_room(room)
        msg = f'{session.type.title()} <{session.name}> joined room <{room}>'
        self.log.info(msg)
        self.__notify_room_join_or_leave(room, msg)

    def __leave_room_and_notify(self, room: str) -> None:
        """
        Leave room and notify other clients in this room.

        Parameters
        ----------
        room : str
            name of the room the client is leaving
        """
        leave_room(room)
        msg = f'{session.name} left room {room}'
        self.log.info(msg)
        self.__notify_room_join_or_leave(room, msg)

    @staticmethod
    def __notify_room_join_or_leave(room: str, msg: str) -> None:
        """
        Notify a room that one of its clients is joining or leaving

        """
        # share message with other users and nodes, except for all_nodes. That
        # room must never be notified for join or leave events since it
        # contains nodes from different collaborations that shouldn't
        # know about each other.
        if room != ALL_NODES_ROOM:
            emit('message', msg, room=room)

    def __alert_node_status(self, online: bool, node: Authenticatable) -> None:
        """
        Send status update of nodes when they change on/offline status

        Parameters
        ----------
        online: bool
            Whether node is coming online or not
        node: Authenticatable
            The node SQLALchemy object
        """
        event = 'node-online' if online else 'node-offline'
        for room in session.rooms:
            self.socketio.emit(
                event,
                {
                    'id': node.id, 'name': node.name,
                    'org_id': node.organization.id
                },
                namespace='/tasks',
                room=room
            )

    @staticmethod
    def __is_identified_client() -> bool:
        """
        Check if client has been identified as an authenticated user or node

        Returns
        -------
        bool
            True if client has been identified, False otherwise
        """
        return hasattr(session, 'auth_id')

    @staticmethod
    def __clean_node_data(node: db.Node) -> None:
        """
        Remove any information from the database that the node shared about
        e.g. its configuration

        Parameters
        ----------
        node: db.Node
            The node SQLALchemy object
        """
        for conf in node.config:
            conf.delete()

    @staticmethod
    def __cleanup() -> None:
        """ Cleanup database connections """
        DatabaseSessionManager.clear_session()



# -*- coding: utf-8 -*-
import logging
import datetime

from flask import request, render_template, g
from flask_jwt_extended import (
    create_access_token,
    decode_token
)
from flask_restful import Api
from jwt.exceptions import DecodeError
from http import HTTPStatus
from sqlalchemy.orm.exc import NoResultFound

from vantage6.common import logger_name, generate_apikey
from vantage6.common.globals import APPNAME
from vantage6.server import db
from vantage6.server.globals import (
    DEFAULT_EMAILED_TOKEN_VALIDITY_MINUTES, DEFAULT_SUPPORT_EMAIL_ADDRESS
)
from vantage6.server.resource import ServicesResources, with_user
from vantage6.server.resource.common.auth_helper import (
    create_qr_uri, user_login
)
from vantage6.server.resource.common.input_schema import (
    ChangePasswordInputSchema,
    RecoverPasswordInputSchema,
    ResetPasswordInputSchema,
    Recover2FAInputSchema,
    Reset2FAInputSchema,
    ResetAPIKeyInputSchema
)

module_name = logger_name(__name__)
log = logging.getLogger(module_name)


def setup(api: Api, api_base: str, services: dict) -> None:
    """
    Setup the recover resource.

    Parameters
    ----------
    api : Api
        Flask restful api instance
    api_base : str
        Base url of the api
    services : dict
        Dictionary with services required for the resource endpoints
    """
    path = "/".join([api_base, module_name])
    log.info(f'Setting up "{path}" and subdirectories')

    api.add_resource(
        ResetPassword,
        path+'/reset',
        endpoint="reset_password",
        methods=('POST',),
        resource_class_kwargs=services
    )

    api.add_resource(
        RecoverPassword,
        path+'/lost',
        endpoint='recover_password',
        methods=('POST',),
        resource_class_kwargs=services
    )

    api.add_resource(
        ResetTwoFactorSecret,
        path+'/2fa/reset',
        endpoint="reset_2fa_secret",
        methods=('POST',),
        resource_class_kwargs=services
    )

    api.add_resource(
        RecoverTwoFactorSecret,
        path+'/2fa/lost',
        endpoint='recover_2fa_secret',
        methods=('POST',),
        resource_class_kwargs=services
    )

    api.add_resource(
        ChangePassword,
        api_base+'/password/change',
        endpoint='change_password',
        methods=('PATCH',),
        resource_class_kwargs=services
    )

    api.add_resource(
        ResetAPIKey,
        path+'/node',
        endpoint="reset_api_key",
        methods=('POST',),
        resource_class_kwargs=services
    )


recover_pw_schema = RecoverPasswordInputSchema()
reset_pw_schema = ResetPasswordInputSchema()
recover_2fa_schema = Recover2FAInputSchema()
reset_2fa_schema = Reset2FAInputSchema()
reset_api_key_schema = ResetAPIKeyInputSchema()
change_pw_schema = ChangePasswordInputSchema()


# ------------------------------------------------------------------------------
# Resources / API's
# ------------------------------------------------------------------------------
class ResetPassword(ServicesResources):
    """User can use recover token to reset their password."""

    def post(self):
        """Set a new password using a recover token
        ---
        description: >-
          User can use a recover token to reset their password

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  reset_token:
                    type: string
                    description: Recover token (received by email)
                  password:
                    type: string
                    description: New password

        responses:
          200:
            description: Ok
          400:
            description: Password or recovery token is missing or invalid

        tags: ["Account recovery"]
        """
        body = request.get_json()
        # validate request body
        errors = reset_pw_schema.validate(body)
        if errors:
            return {'msg': 'Request body is incorrect', 'errors': errors}, \
                HTTPStatus.BAD_REQUEST

        reset_token = body.get("reset_token")
        password = body.get("password")

        # obtain user
        try:
            user_id = decode_token(reset_token)['sub'].get('id')
        except DecodeError:
            return {"msg": "Invalid recovery token!"}, HTTPStatus.BAD_REQUEST

        user = db.User.get(user_id)

        # reset number of failed login attempts to prevent that user cannot
        # reactivate via email
        user.failed_login_attempts = 0
        user.save()

        # set password
        msg = user.set_password(password)
        if msg:
            return {"msg": msg}, HTTPStatus.BAD_REQUEST

        log.info(f"Successfull password reset for '{user.username}'")
        return {"msg": "The password has successfully been reset!"}, \
            HTTPStatus.OK


class RecoverPassword(ServicesResources):
    """Send a mail containing a recover token to reset the password"""

    def post(self):
        """Request a recover token
        ---
        description: >-
          Request a recover token if password is lost. Either email address
          or username must be supplied.

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  username:
                    type: string
                    description: Username from which the password needs to be
                      recovered
                  email:
                    type: string
                    description: Email of user from which the password needs to
                      be recovered

        responses:
          200:
            description: Ok
          400:
            description: No username or email provided

        tags: ["Account recovery"]
        """
        # default return string
        ret = {"msg": "If the username or email is in our database you "
                      "will soon receive an email."}
        body = request.get_json()

        # validate request body
        errors = recover_pw_schema.validate(body)
        if errors:
            return {'msg': 'Request body is incorrect', 'errors': errors}, \
                HTTPStatus.BAD_REQUEST

        # obtain username/email from request
        username = body.get("username")
        email = body.get("email")

        # find user in the database, if not here we stop!
        try:
            if username:
                user = db.User.get_by_username(username)
            else:
                user = db.User.get_by_email(email)
        except NoResultFound:
            account_name = email if email else username
            log.info("Someone request 2FA reset for non-existing account"
                     f" {account_name}")
            # we do not tell them.... But we won't continue either
            return ret

        log.info(f"Password reset requested for '{user.username}'")

        # generate a token that can reset their password
        smtp_settings = self.config.get("smtp", {})
        minutes_token_valid = smtp_settings.get(
            "email_token_validity_minutes",
            DEFAULT_EMAILED_TOKEN_VALIDITY_MINUTES
        )
        expires = datetime.timedelta(minutes=minutes_token_valid)
        reset_token = create_access_token(
            {"id": str(user.id)}, expires_delta=expires
        )

        email_info = self.config.get("smtp", {})
        email_sender = email_info.get("username",
                                      DEFAULT_SUPPORT_EMAIL_ADDRESS)
        support_email = self.config.get("support_email", email_sender)

        self.mail.send_email(
            f"Password reset {APPNAME}",
            sender="support@vantage6.ai",
            recipients=[user.email],
            text_body=render_template(
                "mail/reset_token.txt", token=reset_token,
                firstname=user.firstname, reset_type="password",
                what_to_do="simply ignore this message"
            ),
            html_body=render_template(
                "mail/reset_token.html", token=reset_token,
                firstname=user.firstname, reset_type="password",
                support_email=support_email,
                what_to_do="simply ignore this message"
            )
        )

        return ret


class ResetTwoFactorSecret(ServicesResources):
    """User can use recover token to reset their two-factor authentication."""

    def post(self):
        """Set a new two-factor authentication secret using a recover token
        ---
        description: >-
          User can use a recover token to reset their two-factor authentication
          secret

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  reset_token:
                    type: string
                    description: Recover token (received by email)

        responses:
          200:
            description: Ok
          400:
            description: Recovery token is missing or invalid

        tags: ["Account recovery"]
        """
        # retrieve user based on email or username
        body = request.get_json()

        # validate request body
        errors = reset_2fa_schema.validate(body)
        if errors:
            return {'msg': 'Request body is incorrect', 'errors': errors}, \
                HTTPStatus.BAD_REQUEST

        # obtain user
        reset_token = body.get("reset_token")
        try:
            user_id = decode_token(reset_token)['sub'].get('id')
        except DecodeError:
            return {"msg": "Invalid recovery token!"}, HTTPStatus.BAD_REQUEST

        user = db.User.get(user_id)

        log.info(f"Resetting two-factor authentication for {user.username}")
        return create_qr_uri(user), HTTPStatus.OK


class RecoverTwoFactorSecret(ServicesResources):
    """Send a mail containing a recover token for the 2fa secret"""

    def post(self):
        """Request a recover token to reset two-factor authentication secret
        ---
        description: >-
          Request a recover token if two-factor authentication secret is lost.
          A password and either email address or username must be supplied.

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  username:
                    type: string
                    description: Username from which the 2fa needs to be reset
                  email:
                    type: string
                    description: Email of user from which the 2fa needs to be
                      reset
                  password:
                    type: string
                    description: Password of user whose 2fa needs to be reset

        responses:
          200:
            description: Ok
          400:
            description: No username or email and password provided

        tags: ["Account recovery"]
        """
        # default return string
        ret = {"msg": "If you sent a correct combination of username/email and"
                      "password, you will soon receive an email."}

        # obtain parameters from request
        body = request.get_json()

        # validate request body
        errors = recover_2fa_schema.validate(body)
        if errors:
            return {'msg': 'Request body is incorrect', 'errors': errors}, \
                HTTPStatus.BAD_REQUEST

        username = body.get("username")
        email = body.get("email")
        password = body.get("password")

        # find user in the database, if not here we stop!
        try:
            if username:
                user = db.User.get_by_username(username)
            else:
                user = db.User.get_by_email(email)
        except NoResultFound:
            account_name = email if email else username
            log.info("Someone request 2FA reset for non-existing account"
                     f" {account_name}")
            # we do not tell them.... But we won't continue either
            return ret

        # check password
        user, code = user_login(self.config, user.username, password,
                                self.mail)
        if code != HTTPStatus.OK:
            log.error(f"Failed to reset 2FA for user {username}, wrong "
                      "password")
            return user, code

        log.info(f"2FA reset requested for '{user.username}'")

        # generate a token that can reset their password
        smtp_settings = self.config.get("smtp", {})
        minutes_token_valid = smtp_settings.get(
            "email_token_validity_minutes",
            DEFAULT_EMAILED_TOKEN_VALIDITY_MINUTES
        )
        expires = datetime.timedelta(minutes=minutes_token_valid)
        reset_token = create_access_token(
            {"id": str(user.id)}, expires_delta=expires
        )

        email_info = self.config.get("smtp", {})
        email_sender = email_info.get("username",
                                      DEFAULT_SUPPORT_EMAIL_ADDRESS)
        support_email = self.config.get("support_email", email_sender)

        self.mail.send_email(
            f"Two-factor authentication reset {APPNAME}",
            sender=email_sender,
            recipients=[user.email],
            text_body=render_template(
                "mail/reset_token.txt", token=reset_token,
                firstname=user.firstname,
                reset_type="two-factor authentication code",
                what_to_do=("please reset your password! It has been "
                            "compromised")
            ),
            html_body=render_template(
                "mail/reset_token.html", token=reset_token,
                firstname=user.firstname,
                reset_type="two-factor authentication code",
                support_email=support_email,
                what_to_do=("please reset your password! It has been "
                            "compromised")
            )
        )

        return ret


class ChangePassword(ServicesResources):
    """
    Let user change their password with old password as verification
    """

    @with_user
    def patch(self):
        """Set a new password using the current password
        ---
        description: >-
          Users can change their password by submitting their current password
          and a new password

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  current_password:
                    type: string
                    description: current password
                  new_password:
                    type: string
                    description: new password

        responses:
          200:
            description: Ok
          400:
            description: Current or new password is missing from JSON body, or
              they are the same, or the new password doesn't meet password
              criteria
          401:
            description: Current password is incorrect

        tags: ["Account recovery"]
        """
        body = request.get_json()
        # validate request body
        errors = change_pw_schema.validate(body)
        if errors:
            return {'msg': 'Request body is incorrect', 'errors': errors}, \
                HTTPStatus.BAD_REQUEST

        old_password = body.get("current_password")
        new_password = body.get("new_password")

        user = g.user
        log.debug(f"Changing password for user {user.id}")

        # check if the old password is correct
        pw_correct = user.check_password(old_password)
        if not pw_correct:
            return {"msg": "Your current password is not correct!"}, \
                HTTPStatus.UNAUTHORIZED

        if old_password == new_password:
            return {"msg": "New password is the same as current password!"}, \
                HTTPStatus.BAD_REQUEST

        # set password
        msg = user.set_password(new_password)
        if msg:
            return {"msg": msg}, HTTPStatus.BAD_REQUEST

        log.info(f"Successful password change for '{user.username}'")
        return {"msg": "The password has been changed successfully!"}, \
            HTTPStatus.OK


class ResetAPIKey(ServicesResources):
    """User can reset API key."""

    def __init__(self, socketio, mail, api, permissions, config):
        super().__init__(socketio, mail, api, permissions, config)

        # obtain permissions to check if user is allowed to modify nodes
        self.r = getattr(self.permissions, 'node')

    @with_user
    def post(self):
        """Reset a node's API key
        ---
        description: >-
            If a node's API key is lost, this route can be used to obtain a new
            API key.\n

            ### Permission Table\n
            |Rule name|Scope|Operation|Assigned to node|Assigned to container|
            Description|\n
            |--|--|--|--|--|--|\n
            |Node|Global|Edit|❌|❌|Reset API key of node specified by id|\n
            |Node|Organization|Edit|❌|❌|Reset API key of node specified by
            id which is part of your organization |\n

            Accessible to users.

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  id:
                    type: integer
                    description: ID of node whose API key is to be reset

        responses:
            200:
                description: Ok
            400:
                description: ID missing from json body
            401:
                description: Unauthorized
            404:
                description: Node not found

        security:
            - bearerAuth: []

        tags: ["Account recovery"]
        """
        body = request.get_json()

        # validate request body
        errors = reset_api_key_schema.validate(body)
        if errors:
            return {'msg': 'Request body is incorrect', 'errors': errors}, \
                HTTPStatus.BAD_REQUEST

        id_ = body['id']
        node = db.Node.get(id_)
        if not node:
            return {
                'msg': f'Node id={id_} is not found!'
            }, HTTPStatus.NOT_FOUND

        # check if user is allowed to edit the node
        if not self.r.e_glo.can():
            own = g.user.organization.id == node.organization.id
            if not (self.r.e_org.can() and own):
                return {'msg': 'You lack the permission to do that!'}, \
                    HTTPStatus.UNAUTHORIZED

        # all good, change API key
        log.info(f"Successful API key reset for node {id}")
        api_key = generate_apikey()
        node.api_key = api_key
        node.save()

        return {"api_key": api_key}, HTTPStatus.OK



import datetime

from pathlib import Path

from vantage6.common.globals import APPNAME

#
#   INSTALLATION SETTINGS
#
PACKAGE_FOLDER = Path(__file__).parent.parent.parent

DATA_FOLDER = PACKAGE_FOLDER / APPNAME / "server" / "_data"

SERVER_MODULE_NAME = APPNAME + "-server"

#
#   RUNTIME SETTINGS
#

# Expiretime of JWT tokens
ACCESS_TOKEN_EXPIRES_HOURS = 6

# minimum validity of JWT Tokens in seconds
MIN_TOKEN_VALIDITY_SECONDS = 1800

# Expiration time of refresh tokens
REFRESH_TOKENS_EXPIRE_HOURS = 48

# Minimum time in seconds that a refresh token must be valid *longer than* the
# access token. This is to prevent the access token from expiring before the
# refresh token.
MIN_REFRESH_TOKEN_EXPIRY_DELTA = 1

# Which resources should be initialized. These names correspond to the
# file-names in the resource directory
RESOURCES = ['node', 'collaboration', 'organization', 'task', 'run',
             'token', 'user', 'version', 'recover', 'role',
             'rule', 'health', 'vpn', 'port', 'event']

# Super user information. This user is only created if it is not in the
# database yet at startup time.
SUPER_USER_INFO = {
    "username": "root",
    "password": "root"
}

# default support email address
DEFAULT_SUPPORT_EMAIL_ADDRESS = 'support@vantage6.ai'

# default time that token is valid in minutes
DEFAULT_EMAILED_TOKEN_VALIDITY_MINUTES = 60

# pagination settings
DEFAULT_PAGE = 1
DEFAULT_PAGE_SIZE = 10



import click
import IPython
import yaml

from functools import wraps
from traitlets.config import get_config
from colorama import Fore, Style

from vantage6.common import (
    info,
    error,
)
from vantage6.server.model.base import Database
from vantage6.cli.globals import DEFAULT_SERVER_SYSTEM_FOLDERS as S_FOL
from vantage6.server.controller import fixture
from vantage6.cli.configuration_wizard import (
    select_configuration_questionaire
)
from vantage6.cli.context import ServerContext
from vantage6.server._version import __version__


help_ = {
    "name": "name of the configutation you want to use.",
    "config": "absolute path to configuration-file; overrides NAME",
}


def click_insert_context(func: callable) -> callable:
    """
    Decorator to insert a ServerContext object into the function.

    This decorator will insert a ServerContext object into the function. The
    ServerContext object is created based on the configuration file that is
    selected by the user. The user can select the configuration file by
    supplying the name of the configuration file, or by supplying the path to
    the configuration file. The decorator will also initialize the database
    connection.

    Parameters
    ----------
    func : callable
        The function to decorate.

    Returns
    -------
    callable
        The decorated function.
    """

    # add option decorators
    @click.option('-n', '--name', default=None, help=help_["name"])
    @click.option('-c', '--config', default=None, help=help_["config"])
    @click.option('--system', 'system_folders', flag_value=True)
    @click.option('--user', 'system_folders', flag_value=False, default=S_FOL)
    @wraps(func)
    def func_with_context(name: str, config: str, system_folders: bool,
                          *args, **kwargs) -> callable:

        # select configuration if none supplied
        if config:
            ctx = ServerContext.from_external_config_file(
                config, system_folders
            )
        else:
            if not name:
                try:
                    name = select_configuration_questionaire(
                        "server", system_folders
                    )
                except Exception:
                    error("No configurations could be found!")
                    exit()

            # raise error if config could not be found
            if not ServerContext.config_exists(name, system_folders):
                scope = "system" if system_folders else "user"
                error(
                    f"Configuration {Fore.RED}{name}{Style.RESET_ALL} does not"
                    f" exist in the {Fore.RED}{scope}{Style.RESET_ALL} folder!"
                )
                exit(1)

            # create server context, and initialize db
            ctx = ServerContext(name, system_folders=system_folders)

        # initialize database (singleton)
        allow_drop_all = ctx.config["allow_drop_all"]
        Database().connect(uri=ctx.get_database_uri(),
                           allow_drop_all=allow_drop_all)

        return func(ctx, *args, **kwargs)

    return func_with_context


@click.group(name='server')
def cli_server() -> None:
    """Subcommand `vserver-local`."""
    pass


#
#   import
#
@cli_server.command(name='import')
@click.argument('file_', type=click.Path(exists=True))
@click.option('--drop-all', is_flag=True, default=False)
@click_insert_context
def cli_server_import(ctx: ServerContext, file_: str, drop_all: bool) -> None:
    """
    Import organizations/collaborations/users and tasks. Mainly useful for
    testing purposes.

    Parameters
    ----------
    ctx : ServerContext
        The context of the server instance.
    file_ : str
        The YAML file with resources to import.
    drop_all : bool
        Whether to drop all tables before importing.
    """
    # Note: ctx appears to be unused but is needed for the click_insert_context
    # to select the server in which to import the data.
    info("Reading yaml file.")
    with open(file_) as f:
        entities = yaml.safe_load(f.read())

    info("Adding entities to database.")
    fixture.load(entities, drop_all=drop_all)


#
#   shell
#
@cli_server.command(name='shell')
@click_insert_context
def cli_server_shell(ctx: ServerContext) -> None:
    """
    Run an iPython shell.

    Note that using the shell is not recommended, as there are no checks on
    the validity of the data you are entering. It is better to use the UI,
    Python client, or the API.

    Parameters
    ----------
    ctx : ServerContext
        The context of the server instance.
    """
    # Note: ctx appears to be unused but is needed for the click_insert_context
    # to select the server and start the database connection.
    c = get_config()
    c.InteractiveShellEmbed.colors = "Linux"

    # Suppress logging (e.g. on tab-completion)
    import logging
    logging.getLogger('parso.cache').setLevel(logging.WARNING)
    logging.getLogger('parso.python.diff').setLevel(logging.WARNING)
    logging.getLogger('urllib3.connectionpool').setLevel(logging.WARNING)
    logging.getLogger('asyncio').setLevel(logging.WARNING)

    logging.warning(
        "Using the shell is not recommended! There are no checks on "
        "the validity of the data you are entering.")
    logging.warning("Please use the User interface, Python client, or API.")
    del logging

    import vantage6.server.db as db
    IPython.embed(config=c)


#
#   version
#
@cli_server.command(name='version')
def cli_server_version() -> None:
    """ Prints current version of vantage6 services installed. """
    click.echo(__version__)



# -*- coding: utf-8 -*-
import logging
import base64

from marshmallow import fields
from marshmallow_sqlalchemy import SQLAlchemyAutoSchema
from flask import url_for

from vantage6.server import db
from vantage6.common import logger_name
from vantage6.common.globals import STRING_ENCODING
from vantage6.server.model import Base, User
from vantage6.server.resource.common.pagination import Pagination

log = logging.getLogger(logger_name(__name__))


def create_one_to_many_link(obj: Base, link_to: str, link_from: str) -> str:
    """
    Create an API link to get objects related to a given object.

    Parameters
    ----------
    obj : Base
        Object to which the link is created
    link_to : str
        Name of the resource to which the link is created
    link_from : str
        Name of the resource from which the link is created

    Returns
    -------
    str
        API link

    Examples
    --------
    >>> create_one_to_many_link(obj, "node", "organization_id")
    "/api/node?organization_id=<obj.id>"
    """
    endpoint = link_to + "_without_id"
    values = {link_from: obj.id}
    return url_for(endpoint, **values)


class HATEOASModelSchema(SQLAlchemyAutoSchema):
    """
    This class is used to convert foreign-key fields to HATEOAS specification.
    """

    api = None

    def __init__(self, *args, **kwargs) -> None:

        # set lambda functions to create links for one to one relationship
        setattr(self, "node", lambda obj: self.create_hateoas("node", obj))
        setattr(self, "organization",
                lambda obj: self.create_hateoas("organization", obj))
        setattr(self, "collaboration",
                lambda obj: self.create_hateoas("collaboration", obj))
        setattr(self, "user", lambda obj: self.create_hateoas("user", obj))
        setattr(self, "run", lambda obj: self.create_hateoas("run", obj))
        setattr(self, "task", lambda obj: self.create_hateoas("task", obj))
        setattr(self, "port", lambda obj: self.create_hateoas("port", obj))
        setattr(self, "parent_", lambda obj: self.create_hateoas(
            "parent", obj, endpoint="task"))
        setattr(self, "init_org_", lambda obj: self.create_hateoas(
            "init_org", obj, endpoint="organization"))
        setattr(self, "init_user_", lambda obj: self.create_hateoas(
            "init_user", obj, endpoint="user"))

        # call super class. Do this after setting the attributes above, because
        # the super class initializer will call the attributes.
        super().__init__(*args, **kwargs)

    def create_hateoas(
        self, name: str, obj: Base, endpoint: str = None
    ) -> dict | None:
        """
        Create a HATEOAS link to a related object.

        Parameters
        ----------
        name : str
            Name of the related resource
        obj : Base
            SQLAlchemy resource to which the link is created
        endpoint : str, optional
            Name of the endpoint to which the link is created, by default None.
            If None, the endpoint is assumed to be the same as the name of the
            related resource.

        Returns
        -------
        dict | None
            HATEOAS link to the related object, or None if the related object
            does not exist.
        """
        # get the related object
        elem = getattr(obj, name)

        # create the link
        endpoint = endpoint if endpoint else name
        if elem:
            return self._hateoas_link_from_related_resource(elem, endpoint)
        else:
            return None

    def _hateoas_link_from_related_resource(
        self, elem: Base, name: str
    ) -> dict:
        """
        Construct a HATEOAS link from a related object.

        Parameters
        ----------
        elem : Base
            SQLAlchemy resource for which the link is created
        name : str
            Name of the resource

        Returns
        -------
        dict
            HATEOAS link to the related object
        """
        _id = elem.id
        endpoint = name + "_with_id"
        if self.api:
            if not self.api.owns_endpoint(endpoint):
                Exception(f"Make sure {endpoint} exists!")

            verbs = list(
                self.api.app.url_map._rules_by_endpoint[endpoint][0].methods
            )
            verbs.remove("HEAD")
            verbs.remove("OPTIONS")
            url = url_for(endpoint, id=_id)
            return {"id": _id, "link": url, "methods": verbs}
        else:
            log.error("No API found?")

    def meta_dump(self, pagination: Pagination) -> dict:
        """
        Dump paginated database resources to a dictionary that has links
        to additional pages.

        Parameters
        ----------
        pagination : Pagination
            Paginated database resources

        Returns
        -------
        dict
            Dictionary with paginated database resources and links to
            additional pages.
        """
        data = self.dump(pagination.page.items, many=True)
        return {'data': data, 'links': pagination.metadata_links}


# /task/{id}
class TaskSchema(HATEOASModelSchema):
    class Meta:
        model = db.Task

    status = fields.String()
    finished_at = fields.DateTime()
    collaboration = fields.Method("collaboration")
    runs = fields.Function(lambda obj: create_one_to_many_link(
        obj, link_to="run", link_from="task_id"
    ))
    results = fields.Function(lambda obj: create_one_to_many_link(
        obj, link_to="result", link_from="task_id"
    ))
    parent = fields.Method("parent_")
    children = fields.Function(lambda obj: create_one_to_many_link(
        obj, link_to="task", link_from="parent_id"
    ))
    init_org = fields.Method("init_org_")
    init_user = fields.Method("init_user_")
    databases = fields.Method("databases_")

    @staticmethod
    def databases_(obj):
        return [
            {
                "label": db.database,
                "parameters": db.parameters
            } for db in obj.databases
        ]


class ResultSchema(HATEOASModelSchema):
    class Meta:
        model = db.Run
        exclude = ("assigned_at", "started_at", "finished_at", "status",
                   "ports", "organization", "log", "input",)

    run = fields.Method("make_run_link")
    task = fields.Method("task")

    @staticmethod
    def make_run_link(obj):
        return {
            "id": obj.id,
            "link": url_for("run_with_id", id=obj.id),
            "methods": ["GET", "PATCH"]
        }


# /task/{id}?include=runs
class TaskWithRunSchema(TaskSchema):
    """Returns the TaskSchema plus the correspoding runs."""
    runs = fields.Nested('RunSchema', many=True)


# /task/{id}?include=results
class TaskWithResultSchema(TaskSchema):
    """Returns the TaskSchema plus the correspoding results."""
    results = fields.Nested('ResultSchema', many=True)


# /task/{id}?include=runs,results
class TaskWithRunAndResultSchema(TaskSchema):
    """Returns the TaskSchema plus the correspoding runs and results."""
    runs = fields.Nested('RunSchema', many=True)
    results = fields.Nested('ResultSchema', many=True)




class RunSchema(HATEOASModelSchema):
    class Meta:
        model = db.Run
        exclude = ('result',)

    organization = fields.Method("organization")
    task = fields.Method("task")
    results = fields.Method("result_link")
    node = fields.Function(
        serialize=lambda obj: RunNodeSchema().dump(obj.node, many=False)
    )
    ports = fields.Function(
        serialize=lambda obj: RunPortSchema().dump(obj.ports, many=True)
    )

    @staticmethod
    def result_link(obj):
        return {
            "id": obj.id,
            "link": url_for("result_with_id", id=obj.id),
            "methods": ["GET", "PATCH"]
        }


class RunTaskIncludedSchema(RunSchema):
    task = fields.Nested('TaskSchema', many=False, exclude=["runs"])


class RunNodeSchema(HATEOASModelSchema):
    class Meta:
        model = db.Node
        exclude = ('type', 'api_key', 'collaboration', 'organization',
                   'last_seen')


class PortSchema(HATEOASModelSchema):
    class Meta:
        model = db.AlgorithmPort


class RunPortSchema(HATEOASModelSchema):
    class Meta:
        model = db.AlgorithmPort
        exclude = ('run',)


class OrganizationSchema(HATEOASModelSchema):
    class Meta:
        model = db.Organization
        exclude = ('_public_key',)

    # add links to linked resources
    collaborations = fields.Function(lambda obj: create_one_to_many_link(
        obj, link_to='collaboration', link_from='organization_id'
    ))
    nodes = fields.Function(lambda obj: create_one_to_many_link(
        obj, link_to='node', link_from='organization_id'
    ))
    users = fields.Function(lambda obj: create_one_to_many_link(
        obj, link_to='user', link_from='organization_id'
    ))
    tasks = fields.Function(lambda obj: create_one_to_many_link(
        obj, link_to='task', link_from='init_org_id'
    ))
    runs = fields.Function(lambda obj: create_one_to_many_link(
        obj, link_to='run', link_from='organization_id'
    ))

    public_key = fields.Function(
        lambda obj: (
            base64.b64encode(obj._public_key).decode(STRING_ENCODING)
            if obj._public_key else ""
        )
    )


class CollaborationSchema(HATEOASModelSchema):
    class Meta:
        model = db.Collaboration

    # add links to linked resources
    organizations = fields.Function(lambda obj: create_one_to_many_link(
        obj, link_to='organization', link_from='collaboration_id'
    ))
    nodes = fields.Function(lambda obj: create_one_to_many_link(
        obj, link_to='node', link_from='collaboration_id'
    ))
    tasks = fields.Function(lambda obj: create_one_to_many_link(
        obj, link_to='task', link_from='collaboration_id'
    ))


class NodeSchema(HATEOASModelSchema):
    organization = fields.Method("organization")
    collaboration = fields.Method("collaboration")
    config = fields.Nested('NodeConfigSchema', many=True,
                           exclude=['id'])

    class Meta:
        model = db.Node
        exclude = ('api_key',)


class NodeConfigSchema(HATEOASModelSchema):
    class Meta:
        model = db.NodeConfig


class NodeSchemaSimple(HATEOASModelSchema):
    organization = fields.Method("organization")

    class Meta:
        model = db.Node
        exclude = ('collaboration', 'api_key', 'type',)


class UserSchema(HATEOASModelSchema):

    class Meta:
        model = db.User
        exclude = ('password', 'failed_login_attempts', 'last_login_attempt')

    roles = fields.Function(lambda obj: create_one_to_many_link(
        obj, link_to='role', link_from='user_id'
    ))
    rules = fields.Function(lambda obj: create_one_to_many_link(
        obj, link_to='rule', link_from='user_id'
    ))

    organization = fields.Method("organization")


class UserWithPermissionDetailsSchema(UserSchema):
    """
    A schema for API responses that contains regular user details plus
    additional permission details for the user to be used by the UI.
    """
    permissions = fields.Method("permissions_")

    @staticmethod
    def permissions_(obj: User) -> dict:
        """
        Returns a dictionary containing permission details for the user to be
        used by the UI.

        Parameters
        ----------
        obj : User
            The user to get permission details for.

        Returns
        -------
        dict
            A dictionary containing permission details for the user.
        """
        role_ids = [role.id for role in obj.roles]
        rule_ids = list(set(
            [rule.id for rule in obj.rules] + \
            [rule.id for role in obj.roles for rule in role.rules]
        ))
        # Return which organizations the user is in collaboration with so that
        # UI knows which organizations user has access to if they have
        # collaboration scope permissions
        orgs_in_collabs = list(set([
            org.id
            for collab in obj.organization.collaborations
            for org in collab.organizations
        ]))
        return {
            "roles": role_ids,
            "rules": rule_ids,
            "orgs_in_collabs": orgs_in_collabs
        }


class RoleSchema(HATEOASModelSchema):

    rules = fields.Function(lambda obj: create_one_to_many_link(
        obj, link_to='rule', link_from='role_id'
    ))
    users = fields.Function(lambda obj: create_one_to_many_link(
        obj, link_to='user', link_from='role_id'
    ))
    organization = fields.Method("organization")

    class Meta:
        model = db.Role


class RuleSchema(HATEOASModelSchema):

    scope = fields.Function(serialize=lambda obj: obj.scope.name)
    operation = fields.Function(serialize=lambda obj: obj.operation.name)
    users = fields.Function(lambda obj: create_one_to_many_link(
        obj, link_to='user', link_from='rule_id'
    ))

    class Meta:
        model = db.Rule
        exclude = ('roles',)



class VPNPortalAuthException(Exception):
    """Exception raised when the authentication with the VPN portal fails."""
    pass


class VPNConfigException(Exception):
    """
    Exception raised when the server admin provides invalid VPN configuration.
    """
    pass



# -*- coding: utf-8 -*-
"""
Resources below '/<api_base>/token'
"""
import logging
import pyotp
import json

from flask import request, g
from flask_jwt_extended import (
    jwt_required,
    create_access_token,
    create_refresh_token,
    get_jwt_identity
)
from flask_restful import Api
from http import HTTPStatus

from vantage6 import server
from vantage6.common.task_status import has_task_finished
from vantage6.server import db
from vantage6.server.model.user import User
from vantage6.server.resource import (
    with_node,
    ServicesResources
)
from vantage6.server.resource.common.auth_helper import (
  user_login, create_qr_uri
)
from vantage6.server.resource.common.input_schema import (
    TokenAlgorithmInputSchema,
    TokenNodeInputSchema,
    TokenUserInputSchema
)

module_name = __name__.split('.')[-1]
log = logging.getLogger(module_name)


def setup(api: Api, api_base: str, services: dict) -> None:
    """
    Setup the token resource.

    Parameters
    ----------
    api : Api
        Flask restful api instance
    api_base : str
        Base url of the api
    services : dict
        Dictionary with services required for the resource endpoints
    """
    path = "/".join([api_base, module_name])
    log.info('Setting up "{}" and subdirectories'.format(path))

    api.add_resource(
        UserToken,
        path+'/user',
        endpoint='user_token',
        methods=('POST',),
        resource_class_kwargs=services
    )

    api.add_resource(
        NodeToken,
        path+'/node',
        endpoint='node_token',
        methods=('POST',),
        resource_class_kwargs=services
    )

    api.add_resource(
        ContainerToken,
        path+'/container',
        endpoint='container_token',
        methods=('POST',),
        resource_class_kwargs=services
    )

    api.add_resource(
        RefreshToken,
        path+'/refresh',
        endpoint='refresh_token',
        methods=('POST',),
        resource_class_kwargs=services
    )


user_token_input_schema = TokenUserInputSchema()
node_token_input_schema = TokenNodeInputSchema()
algorithm_token_input_schema = TokenAlgorithmInputSchema()


# ------------------------------------------------------------------------------
# Resources / API's
# ------------------------------------------------------------------------------
class UserToken(ServicesResources):
    """resource for api/token"""

    def post(self):
        """Login user
        ---
        description: >-
          Allow user to sign in by supplying a username and password. When MFA
          is enabled on the server, a code is also required

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  username:
                    type: string
                    description: Username of user that is logging in
                  password:
                    type: string
                    description: User password
                  mfa_code:
                    type: string
                    description: Two-factor authentication code. Only required
                      if two-factor authentication is mandatory.

        responses:
          200:
            description: Ok, authenticated
          400:
            description: Username/password combination unknown, or missing from
              request body.
          401:
            description: Password and/or two-factor authentication token
              incorrect.

        tags: ["Authentication"]
        """
        log.debug("Authenticate user using username and password")

        body = request.get_json()
        # validate request body
        errors = user_token_input_schema.validate(body)
        if errors:
            return {'msg': 'Request body is incorrect', 'errors': errors}, \
                HTTPStatus.BAD_REQUEST

        # Check JSON body
        username = body.get('username')
        password = body.get('password')

        user, code = user_login(self.config, username, password, self.mail)
        if code != HTTPStatus.OK:  # login failed
            log.error(f"Failed to login for user='{username}'")
            return user, code

        is_mfa_enabled = self.config.get('two_factor_auth', False)
        if is_mfa_enabled:
            if user.otp_secret is None:
                # server requires mfa but user hasn't set it up yet. Return
                # an URI to generate a QR code
                log.info(f'Redirecting user {username} to setup MFA')
                return create_qr_uri(user), HTTPStatus.OK
            else:
                # 2nd authentication factor: check the OTP secret of the user
                mfa_code = body.get('mfa_code')
                if not mfa_code:
                    # note: this is not treated as error, but simply guide
                    # user to also fill in second factor
                    return {"msg": "Please include your two-factor "
                            "authentication code"}, HTTPStatus.OK
                elif not self.validate_2fa_token(user, mfa_code):
                    return {
                        "msg": "Your two-factor authentication code is "
                               "incorrect!"
                    }, HTTPStatus.UNAUTHORIZED

        token = _get_token_dict(user, self.api)

        log.info(f"Succesfull login from {username}")
        return token, HTTPStatus.OK, {'jwt-token': token['access_token']}

    @staticmethod
    def validate_2fa_token(user: User, mfa_code: int | str) -> bool:
        """
        Check whether the 6-digit two-factor authentication code is valid

        Parameters
        ----------
        user: User
            The SQLAlchemy model of the user who is authenticating
        mfa_code: int | str
            A six-digit TOTP code from an authenticator app

        Returns
        -------
        bool
          Whether six-digit code is valid or not
        """
        # the option `valid_window=1` means the code from 1 time window (30s)
        # ago, is also valid. This prevents that users around the edge of
        # the time window can still login successfully if server is a bit slow
        return pyotp.TOTP(user.otp_secret).verify(str(mfa_code),
                                                  valid_window=1)


class NodeToken(ServicesResources):

    def post(self):
        """Login node
        ---
        description: >-
          Allows node to sign in using a unique API key. If the login is
          successful this returns a dictionary with access and refresh tokens
          for the node as well as a node_url and a refresh_url.

        requestBody:
          content:
            application/json:
              schema:
                $ref: '#/components/schemas/Node'

        responses:
          200:
            description: Ok, authenticated
          400:
            description: No API key provided in request body.
          401:
            description: Invalid API key

        tags: ["Authentication"]
        """
        log.debug("Authenticate Node using api key")

        body = request.get_json()
        # validate request body
        errors = node_token_input_schema.validate(body)
        if errors:
            return {'msg': 'Request body is incorrect', 'errors': errors}, \
                HTTPStatus.BAD_REQUEST

        # Check JSON body
        api_key = request.json.get('api_key')
        node = db.Node.get_by_api_key(api_key)
        if not node:  # login failed
            log.error("Api key is not recognized")
            return {"msg": "Api key is not recognized!"}, \
                HTTPStatus.UNAUTHORIZED

        token = _get_token_dict(node, self.api)

        log.info(f"Succesfull login as node '{node.id}' ({node.name})")
        return token, HTTPStatus.OK, {'jwt-token': token['access_token']}


class ContainerToken(ServicesResources):

    @with_node
    def post(self):
        """Algorithm container login
        ---
        description: >-
          Generate token for the algorithm container of a specific task.\n

          Not available to users; only for authenticated nodes.

        requestBody:
          content:
            application/json:
              schema:
                $ref: '#/components/schemas/ContainerToken'

        responses:
          200:
            description: Container token generated
          400:
            description: Task does not exist or is already completed
          401:
            description: Key request for invalid image or task

        tags: ["Authentication"]
        """
        log.debug("Creating a token for a container running on a node")

        body = request.get_json()
        # validate request body
        errors = algorithm_token_input_schema.validate(body)
        if errors:
            return {'msg': 'Request body is incorrect', 'errors': errors}, \
                HTTPStatus.BAD_REQUEST

        task_id = body.get("task_id")
        claim_image = body.get("image")

        db_task = db.Task.get(task_id)
        if not db_task:
            log.warning(f"Node {g.node.id} attempts to generate key for task "
                        f"{task_id} that does not exist")
            return {"msg": "Parent task does not exist!"}, \
                HTTPStatus.BAD_REQUEST

        # verify that task the token is requested for exists
        if claim_image != db_task.image:
            log.warning(
                f"Node {g.node.id} attempts to generate key for image "
                f"{claim_image} that does not belong to task {task_id}."
            )
            return {"msg": "Image and task do no match"}, \
                HTTPStatus.UNAUTHORIZED

        # check if the node is in the collaboration to which the task is
        # enlisted
        if g.node.collaboration_id != db_task.collaboration_id:
            log.warning(
                f"Node {g.node.id} attempts to generate key for task {task_id}"
                f" which is outside its collaboration "
                f"({g.node.collaboration_id}/{db_task.collaboration_id})."
            )
            return {"msg": "You are not within the collaboration"}, \
                HTTPStatus.UNAUTHORIZED

        # validate that the task not has been finished yet
        if has_task_finished(db_task.status):
            log.warning(f"Node {g.node.id} attempts to generate a key for "
                        f"completed task {task_id}")
            return {"msg": "Task is already finished!"}, HTTPStatus.BAD_REQUEST

        # container identity consists of its node_id,
        # task_id, collaboration_id and image_id
        container = {
            "client_type": "container",
            "node_id": g.node.id,
            "organization_id": g.node.organization_id,
            "collaboration_id": g.node.collaboration_id,
            "task_id": task_id,
            "image": claim_image,
            "databases": [
                json.loads(db_entry.parameters) | {"label": db_entry.database}
                for db_entry in db_task.databases
            ]
        }
        token = create_access_token(container, expires_delta=False)

        return {'container_token': token}, HTTPStatus.OK


class RefreshToken(ServicesResources):

    @jwt_required(refresh=True)
    def post(self):
        """Refresh token
        ---
        description: >-
          Refresh access token if the previous one is expired.\n

          Your refresh token must be present in the request headers to use
          this endpoint.

        responses:
          200:
            description: Token refreshed

        security:
          - bearerAuth: []

        tags: ["Authentication"]
        """
        user_or_node_id = get_jwt_identity()
        log.info(f'Refreshing token for user or node "{user_or_node_id}"')
        user_or_node = db.Authenticatable.get(user_or_node_id)

        return _get_token_dict(user_or_node, self.api), HTTPStatus.OK


def _get_token_dict(user_or_node: db.Authenticatable, api: Api) -> dict:
    """
    Create a dictionary with the tokens and urls for the user or node.

    Parameters
    ----------
    user_or_node : db.Authenticatable
        The user or node to create the tokens for.
    api : Api
        The api to create the urls for.
    """
    token_dict = {
        'access_token': create_access_token(user_or_node),
        'refresh_token': create_refresh_token(user_or_node),
        'refresh_url': api.url_for(RefreshToken),
    }
    if isinstance(user_or_node, db.User):
        token_dict['user_url'] = api.url_for(server.resource.user.User,
                                             id=user_or_node.id)
    else:
        token_dict['node_url'] = api.url_for(server.resource.node.Node,
                                             id=user_or_node.id)
    return token_dict



from vantage6.server.model.authenticatable import Authenticatable
from vantage6.server.model.user import User
from vantage6.server.model.organization import Organization
from vantage6.server.model.collaboration import Collaboration
from vantage6.server.model.base import Base
from vantage6.server.model.member import Member
from vantage6.server.model.node_config import NodeConfig
from vantage6.server.model.node import Node
from vantage6.server.model.task import Task
from vantage6.server.model.run import Run
from vantage6.server.model.permission import Permission, UserPermission
from vantage6.server.model.role import Role
from vantage6.server.model.rule import Rule
from vantage6.server.model.role_rule_association import role_rule_association
from vantage6.server.model.algorithm_port import AlgorithmPort
from vantage6.server.model.task_database import TaskDatabase



from __future__ import annotations
import bcrypt
import datetime as dt

from sqlalchemy import Column, String, Integer, ForeignKey, DateTime
from sqlalchemy.orm import relationship, validates

from vantage6.server.model.base import DatabaseSessionManager
from vantage6.server.model.authenticatable import Authenticatable
from vantage6.server.model.rule import Operation, Rule, Scope
from vantage6.server.model.common.utils import validate_password


class User(Authenticatable):
    """
    Table to keep track of Users (persons) that can access the system.

    Users always belong to an organization and can have certain
    rights within an organization.

    Attributes
    ----------
    username : str
        Username of the user
    password : str
        Password of the user
    firstname : str
        First name of the user
    lastname : str
        Last name of the user
    email : str
        Email address of the user
    organization_id : int
        Foreign key to the organization to which the user belongs
    failed_login_attempts : int
        Number of failed login attempts
    last_login_attempt : datetime.datetime
        Date and time of the last login attempt
    otp_secret : str
        Secret key for one time passwords
    organization : :class:`~.model.organization.Organization`
        Organization to which the user belongs
    roles : list[:class:`~.model.role.Role`]
        Roles that the user has
    rules : list[:class:`~.model.rule.Rule`]
        Rules that the user has
    created_tasks : list[:class:`~.model.task.Task`]
        Tasks that the user has created
    """
    _hidden_attributes = ['password']

    # overwrite id with linked id to the authenticatable
    id = Column(Integer, ForeignKey('authenticatable.id'), primary_key=True)
    __mapper_args__ = {
        'polymorphic_identity': 'user',
    }

    # fields
    username = Column(String, unique=True)
    password = Column(String)
    firstname = Column(String)
    lastname = Column(String)
    email = Column(String, unique=True)
    organization_id = Column(Integer, ForeignKey("organization.id"))
    failed_login_attempts = Column(Integer, default=0)
    last_login_attempt = Column(DateTime)
    otp_secret = Column(String(32))

    # relationships
    organization = relationship("Organization", back_populates="users")
    roles = relationship("Role", back_populates="users",
                         secondary="Permission")
    rules = relationship("Rule", back_populates="users",
                         secondary="UserPermission")
    created_tasks = relationship("Task", back_populates="init_user")

    def __repr__(self) -> str:
        """
        String representation of the user.

        Returns
        -------
        str
            String representation of the user
        """
        organization = self.organization.name if self.organization else "None"
        return (
            f"<User "
            f"id={self.id}, username='{self.username}', roles='{self.roles}', "
            f"organization='{organization}'"
            f">"
        )

    @validates("password")
    def _validate_password(self, key: str, password: str) -> str:
        """
        Validate the password of the user by hashing it, as it is also hashed
        in the database.

        Parameters
        ----------
        key: str
            Name of the attribute (in this case 'password')
        password: str
            Password of the user

        Returns
        -------
        str
            Hashed password
        """
        return self.hash(password)

    def set_password(self, pw: str) -> str | None:
        """
        Set the password of the current user. This function doesn't save the
        new password to the database

        Parameters
        ----------
        pw: str
            The new password

        Returns
        -------
        str | None
            If the new password fails to pass the checks, a message is
            returned. Else, none is returned
        """
        try:
            validate_password(pw)
        except ValueError as e:
            return str(e)

        self.password = pw
        self.save()

    def check_password(self, pw: str) -> bool:
        """
        Check if the password is correct

        Parameters
        ----------
        pw: str
            Password to check

        Returns
        -------
        bool
            Whether or not the password is correct
        """
        if self.password is not None:
            expected_hash = self.password.encode('utf8')
            return bcrypt.checkpw(pw.encode('utf8'), expected_hash)
        return False

    def is_blocked(self, max_failed_attempts: int,
                   inactivation_in_minutes: int) -> tuple[bool, str | None]:
        """
        Check if user can login or if they are temporarily blocked because they
        entered a wrong password too often

        Parameters
        ----------
        max_failed_attempts: int
            Maximum number of attempts to login before temporary deactivation
        inactivation_minutes: int
            How many minutes an account is deactivated

        Returns
        -------
        bool
            Whether or not user is blocked temporarily
        str | None
            Message if user is blocked, else None
        """
        td_max_blocked = dt.timedelta(minutes=inactivation_in_minutes)
        td_last_login = dt.datetime.now() - self.last_login_attempt \
            if self.last_login_attempt else None
        has_max_attempts = (
            self.failed_login_attempts >= max_failed_attempts
            if self.failed_login_attempts else False
        )
        if has_max_attempts and td_last_login < td_max_blocked:
            minutes_remaining = \
                (td_max_blocked - td_last_login).seconds // 60 + 1
            return True, minutes_remaining
        else:
            return False, None

    @classmethod
    def get_by_username(cls, username: str) -> User:
        """
        Get a user by their username

        Parameters
        ----------
        username: str
            Username of the user

        Returns
        -------
        User
            User with the given username

        Raises
        ------
        NoResultFound
            If no user with the given username exists
        """
        session = DatabaseSessionManager.get_session()
        result = session.query(cls).filter_by(username=username).one()
        session.commit()
        return result

    @classmethod
    def get_by_email(cls, email: str) -> User:
        """
        Get a user by their email

        Parameters
        ----------
        email: str
            Email of the user

        Returns
        -------
        User
            User with the given email

        Raises
        ------
        NoResultFound
            If no user with the given email exists
        """
        session = DatabaseSessionManager.get_session()
        result = session.query(cls).filter_by(email=email).one()
        session.commit()
        return result

    @classmethod
    def username_exists(cls, username) -> bool:
        """
        Check if a user with the given username exists

        Parameters
        ----------
        username: str
            Username to check

        Returns
        -------
        bool
            Whether or not a user with the given username exists
        """
        return cls.exists(field='username', value=username)

    def can(self, resource: str, scope: Scope, operation: Operation) -> bool:
        """
        Check if user is allowed to execute a certain action

        Parameters
        ---------
        resource: str
            The resource type on which the action is to be performed
        scope: Scope
            The scope within which the user wants to perform an action
        operation: Operation
            The operation a user wants to execute

        Returns
        -------
        bool
            Whether or not user is allowed to execute the requested operation
            on the resource
        """
        rule = Rule.get_by_(resource, scope, operation)
        return rule in self.rules or \
            any(rule in role.rules for role in self.roles)



from flask import g

from vantage6.server import db


def obtain_auth_collaborations() -> list[db.Collaboration]:
    """
    Obtain the collaborations that the auth is part of.

    Returns
    -------
    list[db.Collaboration]
        List of collaborations
    """
    if g.user:
        return g.user.organization.collaborations
    elif g.node:
        return g.node.organization.collaborations
    else:
        return [db.Collaboration.get(g.container["collaboration_id"])]


def obtain_auth_organization() -> db.Organization:
    """
    Obtain the organization model from the auth that is logged in.

    Returns
    -------
    db.Organization
        Organization model
    """
    if g.user:
        org_id = g.user.organization.id
    elif g.node:
        org_id = g.node.organization.id
    else:
        org_id = g.container["organization_id"]
    return db.Organization.get(org_id)



import datetime
import logging

from sqlalchemy import Column, Text, DateTime, Integer, ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.orm.exc import NoResultFound, MultipleResultsFound

from vantage6.common import logger_name
from vantage6.server.model.base import Base
from vantage6.server.model import (
    Node,
    Collaboration,
    Organization
)
from vantage6.server.model.task import Task
from vantage6.server.model.base import DatabaseSessionManager

log_ = logging.getLogger(logger_name(__name__))


class Run(Base):
    """
    A Run is the description of a :class:`.~vantage6.server.model.task.Task` as
    executed by a Node.

    The input and result fields will be encrypted and can be only read by the
    intended receiver of the message.

    Attributes
    ----------
    input : str
        Input data of the task
    task_id : int
        Id of the task that was executed
    organization_id : int
        Id of the organization that executed the task
    result : str
        Result of the task
    assigned_at : datetime
        Time when the task was assigned to the node
    started_at : datetime
        Time when the task was started
    finished_at : datetime
        Time when the task was finished
    status : str
        Status of the task
    log : str
        Log of the task
    task : :class:`.~vantage6.server.model.task.Task`
        Task that was executed
    organization : :class:`.~vantage6.server.model.organization.Organization`
        Organization that executed the task
    ports : list[:class:`.~vantage6.server.model.algorithm_port.AlgorithmPort`]
        List of ports that are part of this result
    """

    # fields
    input = Column(Text)
    task_id = Column(Integer, ForeignKey("task.id"))
    organization_id = Column(Integer, ForeignKey("organization.id"))
    result = Column(Text)
    assigned_at = Column(DateTime, default=datetime.datetime.utcnow)
    started_at = Column(DateTime)
    finished_at = Column(DateTime)
    status = Column(Text)
    log = Column(Text)

    # relationships
    task = relationship("Task", back_populates="runs")
    organization = relationship("Organization", back_populates="runs")
    ports = relationship("AlgorithmPort", back_populates="run")

    @property
    def node(self) -> Node:
        """
        Returns the node that is associated with this result.

        Returns
        -------
        model.node.Node
            The node that is associated with this result.
        """
        session = DatabaseSessionManager.get_session()
        try:
            node = session.query(Node)\
                .join(Collaboration)\
                .join(Organization)\
                .join(Run)\
                .join(Task)\
                .filter(Run.id == self.id)\
                .filter(self.organization_id == Node.organization_id)\
                .filter(Task.collaboration_id == Node.collaboration_id)\
                .one()
            session.commit()
        # FIXME 2022-03-03 BvB: the following errors are not currently
        # forwarded to the user as request response. Make that happen.
        except NoResultFound:
            log_.warn("No node exists for organization_id "
                      f"{self.organization_id} in the current collaboration!")
            return None
        except MultipleResultsFound:
            log_.error("Multiple nodes are registered for organization_id "
                       f"{self.organization_id} in the current collaboration. "
                       "Please delete all nodes but one.")
            raise
        return node

    def __repr__(self) -> str:
        """
        Returns a string representation of the result.

        Returns
        -------
        str
            String representation of the result.
        """
        return (
            "<Run "
            f"{self.id}: '{self.task.name}', "
            f"organization: {self.organization.name}, "
            f"collaboration: {self.task.collaboration.name}, "
            f"status: {self.status}"
            ">"
        )



from sqlalchemy import Column, Integer, ForeignKey, String
from sqlalchemy.orm import relationship

from .base import Base


class NodeConfig(Base):
    """
    Table to store node configuration by key-value pairs.

    This information includes e.g. which algorithms are allowed on a certain
    node. The information is stored in the database while the node is
    connected.
    """
    # fields
    node_id = Column('node_id', Integer, ForeignKey('node.id'))
    key = Column(String)
    value = Column(String)

    # relationships
    node = relationship("Node", back_populates="config")



from __future__ import annotations
import vantage6.server.globals as constants

from vantage6.common.context import AppContext
from vantage6.cli.configuration_manager import (
    TestingConfigurationManager
)


class TestContext(AppContext):
    """
    Server context for testing purposes.

    Note that this context is specific to the server: for nodes, there is a
    separate test context.
    """

    INST_CONFIG_MANAGER = TestingConfigurationManager
    LOGGING_ENABLED = False

    @classmethod
    def from_external_config_file(cls) -> TestContext:
        """
        Create a context the unittest configuration file.

        Returns
        -------
        TestContext
            Context object
        """
        return super().from_external_config_file(
            cls.test_config_location(), "unittest", True
        )

    @staticmethod
    def test_config_location() -> str:
        """
        Location of the unittest configuration file.

        Returns
        -------
        str
            Path to the unittest configuration file
        """
        return (constants.PACKAGE_FOLDER / constants.APPNAME /
                "server" / "_data" / "unittest_config.yaml")



import codecs
import os

from os import path
from setuptools import setup, find_namespace_packages
from pathlib import Path

# get current directory
here = Path(path.abspath(path.dirname(__file__)))
parent_dir = here.parent.absolute()

# get the long description from the README file
with codecs.open(path.join(parent_dir, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Read the API version from disk. This file should be located in the package
# folder, since it's also used to set the pkg.__version__ variable.
version_path = os.path.join(here, 'vantage6', 'server', '_version.py')
version_ns = {
    '__file__': version_path
}
with codecs.open(version_path) as f:
    exec(f.read(), {}, version_ns)

# setup the package
setup(
    name='vantage6-server',
    version=version_ns['__version__'],
    description='Vantage6 server',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/vantage6/vantage6',
    packages=find_namespace_packages(),
    python_requires='>=3.10',
    install_requires=[
        'bcrypt==4.0.1',
        'flasgger==0.9.5',
        'flask==2.2.5',
        'Flask-Cors==3.0.10',
        'Flask-JWT-Extended==4.4.4',
        'Flask-Mail==0.9.1',
        'Flask-Principal==0.4.0',
        'Flask-RESTful==0.3.9',
        'flask-marshmallow==0.14.0',
        'Flask-SocketIO==5.3.2',
        'gevent==22.10.2',
        'ipython==8.10.0',
        'kombu==5.2.4',
        'marshmallow==3.19.0',
        'marshmallow-sqlalchemy==0.29.0',
        'PyJWT==2.6.0',
        'pyotp==2.8.0',
        'questionary==1.10.0',
        'requests==2.31.0',
        'requests-oauthlib==1.3.1',
        'schema==0.7.5',
        'SQLAlchemy==1.4.46',
        'werkzeug==2.3.4',
        f'vantage6 == {version_ns["__version__"]}',
        f'vantage6-common == {version_ns["__version__"]}'
    ],
    extras_require={
        'dev': [
            'coverage==6.4.4'
        ]
    },
    package_data={
        'vantage6.server': [
            '__build__',
            '_data/**/*.yaml',
            'server_data/*.yaml',
        ],
    },
    entry_points={
        'console_scripts': [
            'vserver-local=vantage6.server.cli.server:cli_server'
        ]
    }
)



# -*- coding: utf-8 -*-
import logging
import datetime as dt

from http import HTTPStatus
from socket import SocketIO
from flask import request, g
from flask_restful import Api

from vantage6.common import logger_name
from vantage6.common.task_status import has_task_finished, TaskStatus
from vantage6.server.resource import ServicesResources, with_user
from vantage6.server import db
from vantage6.server.permission import (
    Scope,
    Operation,
    PermissionManager
)
from vantage6.server.resource.common.input_schema import (
    KillNodeTasksInputSchema,
    KillTaskInputSchema
)

module_name = logger_name(__name__)
log = logging.getLogger(module_name)


def setup(api: Api, api_base: str, services: dict) -> None:
    """
    Setup the event resource.

    Parameters
    ----------
    api : Api
        Flask restful api instance
    api_base : str
        Base url of the api
    services : dict
        Dictionary with services required for the resource endpoints
    """
    path = "/".join([api_base, module_name])
    log.info(f'Setting up "{path}" and subdirectories')

    api.add_resource(
        KillTask,
        api_base+'/kill/task',
        endpoint='kill_task',
        methods=('POST',),
        resource_class_kwargs=services
    )

    api.add_resource(
        KillNodeTasks,
        api_base+'/kill/node/tasks',
        endpoint='kill_node_tasks',
        methods=('POST',),
        resource_class_kwargs=services
    )


# -----------------------------------------------------------------------------
# Permissions
# -----------------------------------------------------------------------------
def permissions(permissions: PermissionManager) -> None:
    """
    Define the permissions for this resource.

    Parameters
    ----------
    permissions : PermissionManager
        Permission manager instance to which permissions are added
    """
    add = permissions.appender(module_name)

    add(scope=Scope.ORGANIZATION, operation=Operation.RECEIVE,
        description="view websocket events of your organization")
    add(scope=Scope.COLLABORATION, operation=Operation.RECEIVE,
        description="view websocket events of your collaborations")
    add(scope=Scope.GLOBAL, operation=Operation.RECEIVE,
        description="view websocket events")
    add(scope=Scope.ORGANIZATION, operation=Operation.SEND,
        description="send websocket events for your organization")
    add(scope=Scope.COLLABORATION, operation=Operation.SEND,
        description="send websocket events for your collaborations")
    add(scope=Scope.GLOBAL, operation=Operation.SEND,
        description="send websocket events to all collaborations")


kill_task_schema = KillTaskInputSchema()
kill_node_tasks_schema = KillNodeTasksInputSchema()


# ------------------------------------------------------------------------------
# Resources / API's
# ------------------------------------------------------------------------------
class KillTask(ServicesResources):
    """ Provide endpoint to kill all containers executing a certain task """

    def __init__(self, socketio, mail, api, permissions, config):
        super().__init__(socketio, mail, api, permissions, config)
        self.r = getattr(self.permissions, module_name)

    @with_user
    def post(self):
        """Kill task specified by task id
        ---
        description: >-
          Kill a task that is currently being executed on a node.

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Event|Global|Create|❌|❌|Create kill signal for any task|\n
          |Event|Collaboration|Create|❌|❌|
          Create kill signal for any task in your collaboration|\n

          Accessible to users.

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  id:
                    type: integer
                    description: Task id which is to be killed

        responses:
          200:
            description: Ok
          404:
            description: Task not found or already completed
          401:
            description: Unauthorized
          400:
            description: No task id provided

        tags: ["Task"]
        """
        body = request.get_json()

        # validate request body
        errors = kill_task_schema.validate(body)
        if errors:
            return {'msg': 'Request body is incorrect', 'errors': errors}, \
                HTTPStatus.BAD_REQUEST

        id_ = body.get("id")
        task = db.Task.get(id_)
        if not task:
            return {"msg": f"Task id={id_} not found"}, HTTPStatus.NOT_FOUND

        if has_task_finished(task.status):
            return {
                "msg": f"Task {id_} already finished with status "
                       f"'{task.status}', so cannot kill it!"
            }, HTTPStatus.BAD_REQUEST

        # Check permissions. If someone doesn't have global permissions, we
        # check if their organization is part of the appropriate collaboration.
        if not self.r.s_glo.can():
            orgs = task.collaboration.organizations
            if not (self.r.s_col.can() and g.user.organization in orgs):
                return {'msg': 'You lack the permission to do that!'}, \
                    HTTPStatus.UNAUTHORIZED

        # call function to kill the task. This function is outside of the
        # endpoint as it is also used in other endpoints
        kill_task(task, self.socketio)

        return {
            "msg": "Nodes have been instructed to kill any containers running "
                   f"for task {id_}."
        }, HTTPStatus.OK


class KillNodeTasks(ServicesResources):
    """ Provide endpoint to kill all tasks on a certain node """
    def __init__(self, socketio, mail, api, permissions, config):
        super().__init__(socketio, mail, api, permissions, config)
        self.r = getattr(self.permissions, "task")

    @with_user
    def post(self):
        """Kill all tasks on a node specified by node id
        ---
        description: >-
          Kill all tasks that are currently being executed on a certain node.

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Event|Global|Create|❌|❌|Create kill signal for all tasks on any
          node|\n
          |Event|Collaboration|Create|❌|❌|Create kill signal for all tasks
          on any node in collaborations|\n
          |Event|Organization|Create|❌|❌|Create kill signal for all tasks on
          any node of own organization|\n

          Accessible to users.

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  id:
                    type: integer
                    description: Node id on which tasks are to be killed

        responses:
          200:
            description: Ok
          404:
            description: Node not found
          401:
            description: Unauthorized
          400:
            description: No task id provided or node is not online

        tags: ["Task"]
        """
        body = request.get_json()
        # validate request body
        errors = kill_node_tasks_schema.validate(body)
        if errors:
            return {'msg': 'Request body is incorrect', 'errors': errors}, \
                HTTPStatus.BAD_REQUEST

        id_ = body.get("id")
        node = db.Node.get(id_)
        if not node:
            return {"msg": f"Node id={id_} not found"}, HTTPStatus.NOT_FOUND

        if node.status != 'online':
            return {
                "msg": f"Node {id_} is not online so cannot kill its tasks!"
            }, HTTPStatus.BAD_REQUEST

        # Check permissions. If someone doesn't have global permissions, we
        # check if their organization is part of the appropriate collaboration.
        if not self.r.s_glo.can():
            collab_orgs = node.collaboration.organizations
            if not (
                (self.r.s_col.can() and g.user.organization in collab_orgs) or
                (self.r.s_org.can() and
                    node.organization_id == g.user.organization_id)
            ):
                return {'msg': 'You lack the permission to do that!'}, \
                    HTTPStatus.UNAUTHORIZED

        self.socketio.emit(
            'kill_containers', {
                'node_id': node.id,
                'collaboration_id': node.collaboration_id
            },
            namespace='/tasks',
            room=f"collaboration_{node.collaboration_id}",
        )

        return {
            "msg": f"Node {node.id} has been instructed to kill all containers"
                   " running on it."
        }, HTTPStatus.OK


def kill_task(task: db.Task, socket: SocketIO) -> None:
    """
    Send instructions to node(s) to kill a certain task

    Parameters
    ----------
    task: Task
        Task that should be killed
    socket: SocketIO
        SocketIO connection object to communicate kill instructions to node
    """
    # Gather runs and task ids of current task and child tasks
    child_runs = [r for child in task.children for r in child.runs]
    all_runs = task.runs + child_runs
    child_task_ids = [child.id for child in task.children]
    all_task_ids = [task.id] + child_task_ids

    kill_list = [{
        'task_id': task_id,
        'run_id': run.id,
        'organization_id': run.organization_id
    } for run, task_id in zip(all_runs, all_task_ids)]

    # emit socket event to the node to execute the container kills
    socket.emit(
        'kill_containers', {
            'kill_list': kill_list,
            'collaboration_id': task.collaboration.id
        },
        namespace='/tasks',
        room=f"collaboration_{task.collaboration_id}",
    )

    # set tasks and subtasks status to killed
    def set_killed(task: db.Task):
        for run in task.runs:
            run.status = TaskStatus.KILLED
            run.finished_at = dt.datetime.now()
            run.save()

    set_killed(task)
    for subtask in task.children:
        set_killed(subtask)



"""
The Permission table defines which roles have been assigned to which users.
It can contain multiple entries for the same user if they have been assigned
multiple roles.

The UserPermission table defines which extra rules have been assigned to which
users. Apart from roles, users may be assigned extra permissions that allow
them to execute one specific action. This table is used to store those, and
may contain multiple entries for the same user.
"""
from sqlalchemy import Column, Integer, ForeignKey, Table

from .base import Base


Permission = Table(
    'Permission',
    Base.metadata,
    Column('role_id', Integer, ForeignKey('role.id')),
    Column('user_id', Integer, ForeignKey('user.id'))
)

UserPermission = Table(
    'UserPermission',
    Base.metadata,
    Column('rule_id', Integer, ForeignKey('rule.id')),
    Column('user_id', Integer, ForeignKey('user.id'))
)



import logging
import importlib

from collections import namedtuple
from flask_principal import Permission, PermissionDenied

from vantage6.server.globals import RESOURCES
from vantage6.server.default_roles import DefaultRole
from vantage6.server.model.base import Base
from vantage6.server.model.role import Role
from vantage6.server.model.rule import Rule, Operation, Scope
from vantage6.server.model.base import DatabaseSessionManager
from vantage6.server.utils import (
    obtain_auth_collaborations, obtain_auth_organization
)
from vantage6.common import logger_name

module_name = logger_name(__name__)
log = logging.getLogger(module_name)

RuleNeed = namedtuple("RuleNeed", ["name", "scope", "operation"])


# TODO BvB 2023-07-27 this utility is a bit superfluous with the definition
# of the operation and scope enums. We should remove it but then add longer
# values to the enums, which leads to many other changes
def print_operation(operation: Operation) -> str:
    """
    String representation of the operation, that is readable by humans.

    Parameters
    ----------
    operation : Operation
        Operation to be printed

    Returns
    -------
    str
        String representation of the operation

    Raises
    ------
    ValueError
        If the operation is not known
    """
    if operation.VIEW:
        return "view"
    elif operation.EDIT:
        return "edit"
    elif operation.CREATE:
        return "create"
    elif operation.DELETE:
        return "delete"
    elif operation.SEND:
        return "send"
    elif operation.RECEIVE:
        return "receive"
    else:
        raise ValueError(f"Unknown operation {operation}")


def print_scope(scope: Scope) -> str:
    """
    String representation of the scope, that is readable by humans.

    Parameters
    ----------
    scope : Scope
        Scope to be printed

    Returns
    -------
    str
        String representation of the scope

    Raises
    ------
    ValueError
        If the scope is not known
    """
    if scope.ORGANIZATION:
        return "organization"
    elif scope.COLLABORATION:
        return "collaboration"
    elif scope.GLOBAL:
        return "global"
    elif scope.OWN:
        return "own"
    else:
        raise ValueError(f"Unknown scope {scope}")


class RuleCollection(dict):
    """
    Class that tracks a set of all rules for a certain resource name

    Parameters
    ----------
    name: str
        Name of the resource endpoint (e.g. node, organization, user)
    """

    def __init__(self, name: str) -> None:
        self.name = name

    def add(self, scope: Scope, operation: Operation) -> None:
        """
        Add a rule to the rule collection

        Parameters
        ----------
        scope: Scope
            Scope within which to apply the rule
        operation: Operation
            What operation the rule applies to
        """
        permission = Permission(RuleNeed(self.name, scope, operation))
        self.__setattr__(f'{operation}_{scope}', permission)

    def can_for_org(self, operation: Operation,
                    subject_org_id: int | str) -> bool:
        """
        Check if an operation is allowed on a certain organization

        Parameters
        ----------
        operation: Operation
            Operation to check if allowed
        subject_org_id: int | str
            Organization id on which the operation should be allowed. If a
            string is given, it will be converted to an int

        Returns
        -------
        bool
            True if the operation is allowed on the organization, False
            otherwise
        """
        if isinstance(subject_org_id, str):
            subject_org_id = int(subject_org_id)

        auth_org = obtain_auth_organization()

        # check if the entity has global permission
        global_perm = getattr(self, f'{operation}_{Scope.GLOBAL}')
        if global_perm and global_perm.can():
            return True

        # check if the entity has organization permission and organization is
        # the same as the subject organization
        org_perm = getattr(self, f'{operation}_{Scope.ORGANIZATION}')
        if auth_org.id == subject_org_id and org_perm and org_perm.can():
            return True

        # check if the entity has collaboration permission and the subject
        # organization is in the collaboration of the own organization
        col_perm = getattr(self, f'{operation}_{Scope.COLLABORATION}')
        if col_perm and col_perm.can():
            for col in auth_org.collaborations:
                if subject_org_id in [org.id for org in col.organizations]:
                    return True

        # no permission found
        return False

    def can_for_col(self, operation: Operation,
                    collaboration_id: int | str) -> bool:
        """
        Check if the user or node can perform the operation on a certain
        collaboration

        Parameters
        ----------
        operation: Operation
            Operation to check if allowed
        collaboration_id: int | str
            Collaboration id on which the operation should be allowed. If a
            string is given, it will be converted to an int
        """
        if isinstance(collaboration_id, str):
            collaboration_id = int(collaboration_id)

        auth_collabs = obtain_auth_collaborations()

        # check if the entity has global permission
        global_perm = getattr(self, f'{operation}_{Scope.GLOBAL}')
        if global_perm and global_perm.can():
            return True

        # check if the entity has collaboration permission and the subject
        # collaboration is in the collaborations of the user/node
        col_perm = getattr(self, f'{operation}_{Scope.COLLABORATION}')
        if col_perm and col_perm.can() and \
                self._id_in_list(collaboration_id, auth_collabs):
            return True

        # no permission found
        return False

    def get_max_scope(self, operation: Operation) -> Scope | None:
        """
        Get the highest scope that the entity has for a certain operation

        Parameters
        ----------
        operation: Operation
            Operation to check

        Returns
        -------
        Scope | None
            Highest scope that the entity has for the operation. None if the
            entity has no permission for the operation
        """
        if getattr(self, f'{operation}_{Scope.GLOBAL}'):
            return Scope.GLOBAL
        elif getattr(self, f'{operation}_{Scope.COLLABORATION}'):
            return Scope.COLLABORATION
        elif getattr(self, f'{operation}_{Scope.ORGANIZATION}'):
            return Scope.ORGANIZATION
        elif getattr(self, f'{operation}_{Scope.OWN}'):
            return Scope.OWN
        else:
            return None

    def has_at_least_scope(self, scope: Scope, operation: Operation) -> bool:
        """
        Check if the entity has at least a certain scope for a certain
        operation

        Parameters
        ----------
        scope: Scope
            Scope to check if the entity has at least
        operation: Operation
            Operation to check

        Returns
        -------
        bool
            True if the entity has at least the scope, False otherwise
        """
        scopes: list[Scope] = self._get_scopes_from(scope)
        for s in scopes:
            perm = getattr(self, f'{operation}_{s}')
            if perm and perm.can():
                return True
        return False

    def _id_in_list(self, id_: int, resource_list: list[Base]) -> bool:
        """
        Check if resource list contains a resource with a certain ID

        Parameters
        ----------
        id_ : int
            ID of the resource
        resource_list : list[db.Base]
            List of resources

        Returns
        -------
        bool
            True if resource is in list, False otherwise
        """
        return any(r.id == id_ for r in resource_list)

    def _get_scopes_from(self, minimal_scope: Scope) -> list[Scope]:
        """
        Get scopes that are at least equal to a certain scope

        Parameters
        ----------
        minimal_scope: Scope
            Minimal scope

        Returns
        -------
        list[Scope]
            List of scopes that are at least equal to the minimal scope

        Raises
        ------
        ValueError
            If the minimal scope is not known
        """
        if minimal_scope == Scope.ORGANIZATION:
            return [Scope.ORGANIZATION, Scope.COLLABORATION, Scope.GLOBAL]
        elif minimal_scope == Scope.COLLABORATION:
            return [Scope.COLLABORATION, Scope.GLOBAL]
        elif minimal_scope == Scope.GLOBAL:
            return [Scope.GLOBAL]
        elif minimal_scope == Scope.OWN:
            return [Scope.OWN, Scope.ORGANIZATION, Scope.COLLABORATION,
                    Scope.GLOBAL]
        else:
            raise ValueError(f"Unknown scope '{minimal_scope}'")


class PermissionManager:
    """
    Loads the permissions and syncs rules in database with rules defined in
    the code
    """

    def __init__(self) -> None:
        self.collections: dict[str, RuleCollection] = {}
        log.info("Loading permission system...")
        self.load_rules_from_resources()

    def load_rules_from_resources(self) -> None:
        """
        Collect all permission rules from all registered API resources
        """
        for res in RESOURCES:
            module = importlib.import_module('vantage6.server.resource.' + res)
            try:
                module.permissions(self)
            except Exception:
                module_name = module.__name__.split(".")[-1]
                log.debug(f"Resource '{module_name}' contains no or invalid "
                          "permissions")

    def assign_rule_to_root(self, name: str, scope: Scope,
                            operation: Operation) -> None:
        """
        Assign a rule to the root role.

        resource: str
            Resource that the rule applies to
        scope: Scope
            Scope that the rule applies to
        operation: Operation
            Operation that the rule applies to
        """
        self.assign_rule_to_fixed_role(DefaultRole.ROOT, name, scope,
                                       operation)

    def assign_rule_to_node(self, resource: str, scope: Scope,
                            operation: Operation) -> None:
        """
        Assign a rule to the Node role.

        Parameters
        ----------
        resource: str
            Resource that the rule applies to
        scope: Scope
            Scope that the rule applies to
        operation: Operation
            Operation that the rule applies to
        """
        self.assign_rule_to_fixed_role(DefaultRole.NODE, resource, scope,
                                       operation)

    def assign_rule_to_container(self, resource: str, scope: Scope,
                                 operation: Operation) -> None:
        """
        Assign a rule to the container role.

        Parameters
        ----------
        resource: str
            Resource that the rule applies to
        scope: Scope
            Scope that the rule applies to
        operation: Operation
            Operation that the rule applies to
        """
        self.assign_rule_to_fixed_role(DefaultRole.CONTAINER, resource, scope,
                                       operation)

    @staticmethod
    def assign_rule_to_fixed_role(fixedrole: str, resource: str, scope: Scope,
                                  operation: Operation) -> None:
        """
        Attach a rule to a fixed role (not adjustable by users).

        Parameters
        ----------
        fixedrole: str
            Name of the fixed role that the rule should be added to
        resource: str
            Resource that the rule applies to
        scope: Scope
            Scope that the rule applies to
        operation: Operation
            Operation that the rule applies to
        """
        role = Role.get_by_name(fixedrole)
        if not role:
            log.warning(f"{fixedrole} role not found, creating it now!")
            role = Role(name=fixedrole, description=f"{fixedrole} role")

        rule = Rule.get_by_(resource, scope, operation)
        if not rule:
            log.error(f"Rule ({resource},{scope},{operation}) not found!")

        if rule not in role.rules:
            role.rules.append(rule)
            log.info(f"Rule ({resource},{scope},{operation}) added to "
                     f"{fixedrole} role!")

    def register_rule(self, resource: str, scope: Scope,
                      operation: Operation, description=None,
                      assign_to_node=False, assign_to_container=False) -> None:
        """
        Register a permission rule in the database.

        If a rule already exists, nothing is done. This rule can be used in API
        endpoints to determine if a user, node or container can do a certain
        operation in a certain scope.

        Parameters
        ----------
        resource : str
            API resource that the rule applies to
        scope : Scope
            Scope of the rule
        operation : Operation
            Operation of the rule
        description : String, optional
            Human readable description where the rule is used for, by default
            None
        assign_to_node: bool, optional
            Whether rule should be assigned to the node role or not. Default
            False
        assign_to_container: bool, optional
            Whether rule should be assigned to the container role or not.
            Default False
        """
        # verify that the rule is in the DB, so that these can be assigned to
        # roles and users
        rule = Rule.get_by_(resource, scope, operation)
        if not rule:
            rule = Rule(name=resource, operation=operation, scope=scope,
                        description=description)
            rule.save()
            log.debug(f"New auth rule '{resource}' with scope={scope}"
                      f" and operation={operation} is stored in the DB")

        if assign_to_container:
            self.assign_rule_to_container(resource, scope, operation)

        if assign_to_node:
            self.assign_rule_to_node(resource, scope, operation)

        # assign all new rules to root user
        self.assign_rule_to_root(resource, scope, operation)

        self.collection(resource).add(rule.scope, rule.operation)

    def appender(self, name: str) -> callable:
        """
        Add a module's rules to the rule collection

        Parameters
        ----------
        name: str
            The name of the module whose rules are to be registered

        Returns
        -------
        Callable
            A callable ``register_rule`` function
        """
        # make sure collection exists
        self.collection(name)
        return lambda *args, **kwargs: self.register_rule(name, *args,
                                                          **kwargs)

    def collection(self, name: str) -> RuleCollection:
        """
        Get a RuleCollection object. If it doesn't exist yet, it will be
        created.

        Parameters
        ----------
        name: str
            Name of the module whose RuleCollection is to be obtained or
            created

        Returns
        -------
        RuleCollection
            The collection of rules belonging to the module name
        """
        if self._collection_exists(name):
            return self.collections[name]
        else:
            self.collections[name] = RuleCollection(name)
            return self.collections[name]

    def _collection_exists(self, name: str) -> bool:
        """
        Check if a module's rule collection is defined

        Parameters
        ----------
        name: str
            Name of the module to be checked

        Returns
        -------
        bool:
            True if RuleCollection is defined for module, else False
        """
        return name in self.collections

    def __getattr__(self, name: str) -> RuleCollection:
        # TODO BvB 2023-01-18 I think this function might not be used. It would
        # be triggered when we do something like
        # `permissionManager.resource_name` but we don't ever do that (?!)
        try:
            collection = self.collections[name]
            return collection
        except Exception as e:
            log.critical(f"Missing permission collection! {name}")
            raise e

    @staticmethod
    def rule_exists_in_db(name: str, scope: Scope,
                          operation: Operation) -> bool:
        """Check if the rule exists in the DB.

        Parameters
        ----------
        name: str
            Name of the rule
        scope: Scope
            Scope of the rule
        operation: Operation
            Operation of the rule

        Returns
        -------
        bool
            Whenever this rule exists in the database or not
        """
        session = DatabaseSessionManager.get_session()
        result = session.query(Rule).filter_by(
            name=name,
            operation=operation,
            scope=scope
        ).scalar()
        session.commit()
        return result

    def check_user_rules(self, rules: list[Rule]) -> dict | None:
        """
        Check if a user, node or container has all the `rules` in a list

        Parameters
        ----------
        rules: list[:class:`~vantage6.server.model.rule.Rule`]
            List of rules that user is checked to have

        Returns
        -------
        dict | None
            Dict with a message which rule is missing, else None
        """
        for rule in rules:
            if not self.collections[rule.name].has_at_least_scope(
                rule.scope, rule.operation
            ):
                return {"msg": f"You don't have the rule ({rule.name}, "
                        f"{print_scope(rule.scope)}, "
                        f"{print_operation(rule.operation)})"}
        return None
